from xdsl import ensure_mlir_module_loaded

ensure_mlir_module_loaded()

from xdsl import _mlir_module as mlir  # type: ignore

from xdsl.ir import (MLContext, SSAValue, OpResult, Block, Operation, Region,
                     BlockArgument, Attribute)
from xdsl.dialects.builtin import (DenseIntOrFPElementsAttr, IntegerAttr,
                                   VectorType, IntegerType, IndexType,
                                   ArrayAttr, FlatSymbolRefAttr, StringAttr,
                                   FunctionType, TupleType, ModuleOp,
                                   Float32Type, SymbolNameAttr, UnitAttr)
from xdsl.dialects.memref import MemRefType
from xdsl.dialects.llvm import LLVMStructType
from typing import Dict


class MLIRConverter:

    def __init__(self, ctx: MLContext):
        self.ctx = ctx
        self.op_to_mlir_ops: Dict[Operation, mlir.ir.Operation] = dict()
        self.block_to_mlir_blocks: Dict[Block, mlir.ir.Block] = dict()

    def register_external_dialects(self):
        pass

    def convert_function_type(self, typ: FunctionType) -> mlir.ir.FunctionType:
        input_array = typ.parameters[0]
        output_array = typ.parameters[1]
        inputs = [
            self.convert_type(input_type_attr)
            for input_type_attr in input_array.data
        ]
        outputs = [
            self.convert_type(output_type_attr)
            for output_type_attr in output_array.data
        ]
        return mlir.ir.FunctionType.get(inputs, outputs)

    def get_mlir_struct_string(self, typ: LLVMStructType) -> str:
        types = []
        for t in typ.types.data:
            types.append(
                self.get_mlir_struct_string(t) if isinstance(
                    t, LLVMStructType) else str(self.convert_type(t)))
        type_string = "(" + ", ".join(types) + ")"

        if typ.struct_name.data == "":
            return "!llvm.struct<" + type_string + ">"
        return "!llvm.struct<" + typ.struct_name.data + ", " + type_string + ">"

    def convert_type(self, typ: Attribute) -> mlir.ir.Type:
        if isinstance(typ, Float32Type):
            return mlir.ir.F32Type.get()
        if isinstance(typ, IntegerType):
            return mlir.ir.IntegerType.get_signless(typ.width.data)
        if isinstance(typ, IndexType):
            return mlir.ir.IndexType.get()
        if isinstance(typ, FunctionType):
            return self.convert_function_type(typ)
        if isinstance(typ, MemRefType):
            return mlir.ir.MemRefType.get(typ.get_shape(),
                                          self.convert_type(typ.element_type))
        if isinstance(typ, TupleType):
            return mlir.ir.TupleType.get_tuple(
                [self.convert_type(t) for t in typ.types.data])
        if isinstance(typ, LLVMStructType):
            mlir_module = mlir.ir.Module.parse(
                "%0 = llvm.mlir.undef : " + self.get_mlir_struct_string(typ))
            return mlir_module.body.operations.__getitem__(
                0).operation.result.type
        raise Exception(f"Unsupported type for mlir conversion: {typ}")

    def convert_value(self, value: SSAValue) -> mlir.ir.Value:
        if isinstance(value, OpResult):
            mlir_op = self.op_to_mlir_ops[value.op]
            return mlir_op.results[value.result_index]
        elif isinstance(value, BlockArgument):
            mlir_block = self.block_to_mlir_blocks[value.block]
            return mlir_block.arguments[value.index]
        raise Exception("Unknown value")

    def convert_attribute(self, attr: Attribute) -> mlir.ir.Attribute:
        if isinstance(attr, StringAttr):
            return mlir.ir.StringAttr.get(attr.data)
        if isinstance(attr, IntegerAttr):
            return mlir.ir.IntegerAttr.get(
                self.convert_type(attr.parameters[1]), attr.parameters[0].data)
        if isinstance(attr, ArrayAttr):
            return mlir.ir.ArrayAttr.get(
                [self.convert_attribute(sub_attr) for sub_attr in attr.data])
        if isinstance(attr, DenseIntOrFPElementsAttr):
            # TODO fix this as soon as DEneIntElementsAttr allow us to pass
            #   vector or tensor
            assert (attr.type.get_num_dims() == 1)
            element_type = self.convert_type(attr.type.element_type)
            typ = "vector" if isinstance(attr.type, VectorType) else "tensor"

            return mlir.ir.DenseIntElementsAttr.parse(
                f"dense<{[d.parameters[0].data for d in attr.data.data]}> : "
                f"{typ}<{len(attr.data.data)}x{element_type}>")
        if isinstance(attr, FlatSymbolRefAttr):
            return mlir.ir.FlatSymbolRefAttr.get(attr.parameters[0].data)
        if isinstance(attr, UnitAttr):
            return mlir.ir.UnitAttr.get()
        # SymbolNameAttrs are in fact just StringAttrs
        if isinstance(attr, SymbolNameAttr):
            return mlir.ir.StringAttr.get(attr.parameters[0].data)
        try:
            return mlir.ir.TypeAttr.get(self.convert_type(attr))
        except Exception:
            raise Exception(
                f"Unsupported attribute for mlir conversion: {attr}")

    def convert_op(self, op: Operation) -> mlir.ir.Operation:
        result_types = [self.convert_type(result.typ) for result in op.results]
        operands = [self.convert_value(operand) for operand in op.operands]
        attributes = {
            name: self.convert_attribute(attr)
            for name, attr in op.attributes.items()
        }

        successors = [
            self.block_to_mlir_blocks[succ] for succ in op.successors
        ]

        mlir_op = mlir.ir.Operation.create(op.name,
                                           results=result_types,
                                           operands=operands,
                                           attributes=attributes,
                                           successors=successors,
                                           regions=len(op.regions))
        self.op_to_mlir_ops[op] = mlir_op

        for region_idx in range(len(op.regions)):
            self.convert_region(op.regions[region_idx],
                                mlir_op.regions[region_idx])
        return mlir_op

    def convert_block(self, block: Block, mlir_block: mlir.ir.Block) -> None:
        ip = mlir.ir.InsertionPoint.at_block_begin(mlir_block)
        for op in block.ops:
            ip.insert(self.convert_op(op))
        return mlir_block

    def convert_region(self, region: Region,
                       mlir_region: mlir.ir.Region) -> None:
        assert (len(mlir_region.blocks) == 0)
        for block in region.blocks:
            mlir_block_args = [
                self.convert_type(block_operand.typ)
                for block_operand in block.args
            ]
            mlir_block = mlir_region.blocks.append(*mlir_block_args)
            self.block_to_mlir_blocks[block] = mlir_block

        for block in region.blocks:
            mlir_block = self.block_to_mlir_blocks[block]
            self.convert_block(block, mlir_block)

    def convert_module_with_ctx(self, op: Operation,
                                mlir_ctx: mlir.ir.Context) -> mlir.ir.Module:
        with mlir.ir.Location.unknown(mlir_ctx):
            if not isinstance(op, ModuleOp):
                raise Exception("top-level operation should be a ModuleOp")
            mlir_module = mlir.ir.Module.create()
            mlir_block = mlir_module.operation.regions[0].blocks[0]
            block = op.regions[0].blocks[0]

            ip = mlir.ir.InsertionPoint.at_block_begin(mlir_block)
            for op in block.ops:
                ip.insert(self.convert_op(op))
            return mlir_module

    def convert_module(self, op: Operation) -> mlir.ir.Module:
        with mlir.ir.Context() as mlir_ctx:
            mlir_ctx.allow_unregistered_dialects = True
            self.register_external_dialects()
            return self.convert_module_with_ctx(op, mlir_ctx)



import pytest

docutils = pytest.importorskip("mlir")

from xdsl.mlir_converter import MLIRConverter, mlir
from xdsl.dialects.scf import Scf
from xdsl.dialects.func import Func
from xdsl.dialects.memref import MemRef
from xdsl.dialects.affine import Affine
from xdsl.dialects.arith import Arith

from xdsl.parser import XDSLParser
from xdsl.ir import MLContext
from xdsl.dialects.builtin import Builtin


def convert_and_verify(test_prog: str):
    ctx = MLContext()
    ctx.register_dialect(Builtin)
    ctx.register_dialect(Func)
    ctx.register_dialect(Affine)
    ctx.register_dialect(Arith)
    ctx.register_dialect(Scf)
    ctx.register_dialect(MemRef)

    parser = XDSLParser(ctx, test_prog)
    module = parser.parse_op()
    module.verify()

    converter = MLIRConverter(ctx)
    with mlir.ir.Context() as mlir_ctx:
        mlir_ctx.allow_unregistered_dialects = True
        with mlir.ir.Location.unknown(mlir_ctx):
            mlir_prog = converter.convert_op(module)
            print(mlir_prog)
            assert (mlir_prog.verify())


def test_func_conversion():
    test_prog = """
builtin.module() {
  func.func() ["sym_name" = "test", "function_type" = !fun<[], []>, "sym_visibility" = "private"]
  {
    func.return()
  }
  func.call() ["callee" = !flat_symbol_ref<"test">]
}
    """
    convert_and_verify(test_prog)


def test_arith_conversion():
    test_prog = """
builtin.module() {
  func.func() ["sym_name" = "test", "function_type" = !fun<[!i32], [!i32]>, "sym_visibility" = "private"]{
  ^0(%arg : !i32):
    %0 : !i32 = arith.constant() ["value" = 0 : !i32]
    %res : !i32 = arith.addi(%arg : !i32, %0 : !i32)
    func.return(%res : !i32)
  }
}
    """
    convert_and_verify(test_prog)


def test_scf_conversion():
    test_prog = """
builtin.module() {
  func.func() ["sym_name" = "test", "function_type" = !fun<[!i32], [!i32]>, "sym_visibility" = "private"]{
  ^0(%arg : !i32):
    %0 : !i32 = arith.constant() ["value" = 42 : !i32]
    %t : !i1 = arith.cmpi(%arg : !i32, %0 : !i32) ["predicate" = 1 : !i64]
    %res : !i32 = scf.if(%t : !i1) {
        %then : !i32 = arith.addi(%arg : !i32, %0 : !i32)
        scf.yield(%then : !i32)
    } {
        %else : !i32 = arith.subi(%arg : !i32, %0 : !i32)
        scf.yield(%else : !i32)
    }
    func.return(%res : !i32)
  }
}
    """
    convert_and_verify(test_prog)


def test_variadic_conversion():
    test_prog = """
builtin.module() {
  func.func() ["sym_name" = "test", "function_type" = !fun<[], []>, "sym_visibility" = "private"]
  {
    %1 : !memref<[1 : !i64], !i32> = memref.alloca() ["alignment" = 0 : !i64, "operand_segment_sizes" = !dense<!vector<[2 : !i64], !i32>, [0 : !i32, 0 : !i32]>]

    func.return()
  }
}
    """
    convert_and_verify(test_prog)


def test_memref_conversion():
    test_prog = """
builtin.module() {
  func.func() ["sym_name" = "sum", "function_type" = !fun<[!i32, !i32], [!i32]>, "sym_visibility" = "public"]{
  ^0(%0 : !i32, %1 : !i32):
    %2 : !index = arith.constant() ["value" = 0 : !index]
    %3 : !memref<[1 : !index], !i32> = memref.alloca() ["alignment" = 0 : !i64, "operand_segment_sizes" = !dense<!vector<[2 : !index], !i32>, [0 : !i32, 0 : !i32]>]
    memref.store(%0 : !i32, %3 : !memref<[1 : !index], !i32>, %2 : !index)
    %4 : !index = arith.constant() ["value" = 0 : !index]
    %5 : !memref<[1 : !index], !i32> = memref.alloca() ["alignment" = 0 : !i64, "operand_segment_sizes" = !dense<!vector<[2 : !index], !i32>, [0 : !i32, 0 : !i32]>]
    memref.store(%1 : !i32, %5 : !memref<[1 : !index], !i32>, %4 : !index)
    %6 : !index = arith.constant() ["value" = 0 : !index]
    %7 : !i32 = memref.load(%3 : !memref<[1 : !index], !i32>, %6 : !index)
    %8 : !index = arith.constant() ["value" = 0 : !index]
    %9 : !i32 = memref.load(%5 : !memref<[1 : !index], !i32>, %8 : !index)
    %10 : !i32 = arith.addi(%7 : !i32, %9 : !i32)
    func.return(%10 : !i32)
  }
}
    """
    convert_and_verify(test_prog)


def test_unit_attr_conversion():
    test_prog = """
builtin.module() {
  func.func() ["sym_name" = "test", "function_type" = !fun<[], []>, "sym_visibility" = "private", "llvm.emit_c_interface"]
  {
    %1 : !memref<[1 : !i64], !i32> = memref.alloca() ["alignment" = 0 : !i64, "operand_segment_sizes" = !dense<!vector<[2 : !i64], !i32>, [0 : !i32, 0 : !i32]>]

    func.return()
  }
}
    """
    convert_and_verify(test_prog)


# NOTE we currently do not test affine conversion because the dialect is incomplete



import argparse
import sys
import os
from io import IOBase, StringIO
import coverage
from typing.io import IO

from xdsl.ir import MLContext
from xdsl.parser import XDSLParser, MLIRParser
from xdsl.printer import Printer
from xdsl.dialects.func import Func
from xdsl.dialects.scf import Scf
from xdsl.dialects.affine import Affine
from xdsl.dialects.arith import Arith
from xdsl.dialects.builtin import ModuleOp, Builtin
from xdsl.dialects.cmath import CMath
from xdsl.dialects.cf import Cf
from xdsl.dialects.vector import Vector
from xdsl.dialects.memref import MemRef
from xdsl.dialects.llvm import LLVM
from xdsl.dialects.irdl import IRDL

from xdsl.irdl_mlir_printer import IRDLPrinter
from xdsl.utils.exceptions import DiagnosticException

from typing import Dict, Callable, List


class xDSLOptMain:
    ctx: MLContext
    args: argparse.Namespace
    """
    The argument parsers namespace which holds the parsed commandline
    attributes.
    """

    available_frontends: Dict[str, Callable[[IOBase], ModuleOp]]
    """
    A mapping from file extension to a frontend that can handle this
    file type.
    """

    available_passes: Dict[str, Callable[[MLContext, ModuleOp], None]]
    """
    A mapping from pass names to functions that apply the pass to a  ModuleOp.
    """

    available_targets: Dict[str, Callable[[ModuleOp, IOBase], None]]
    """
    A mapping from target names to functions that serialize a ModuleOp into a
    stream.
    """

    pipeline: List[tuple[str, Callable[[ModuleOp], None]]]
    """ The pass-pipeline to be applied. """

    def __init__(self,
                 description: str = 'xDSL modular optimizer driver',
                 args=None):
        self.available_frontends = {}
        self.available_passes = {}
        self.available_targets = {}

        self.ctx = MLContext()
        self.register_all_dialects()
        self.register_all_frontends()
        self.register_all_passes()
        self.register_all_targets()

        # arg handling
        arg_parser = argparse.ArgumentParser(description=description)
        self.register_all_arguments(arg_parser)
        self.args = arg_parser.parse_args(args=args)

        self.setup_pipeline()

    def run(self):
        """
        Executes the different steps.
        """
        if self.args.generate_coverage:
            if self.args.exec_root:
                os.chdir(self.args.exec_root)
            cov = coverage.Coverage(config_file='.coveragerc',
                                    auto_data=True,
                                    data_file='.coverage',
                                    data_suffix=True)

            cov.start()

        module = self.parse_input()
        if not self.args.verify_diagnostics:
            self.apply_passes(module)
        else:
            try:
                self.apply_passes(module)
            except DiagnosticException as e:
                print(e)
                exit(0)

        contents = self.output_resulting_program(module)
        self.print_to_output_stream(contents)

        if self.args.generate_coverage:
            cov.stop()

    def register_all_arguments(self, arg_parser: argparse.ArgumentParser):
        """
        Registers all the command line arguments that are used by this tool.

        Add other/additional arguments by overloading this function.
        """
        arg_parser.add_argument("input_file",
                                type=str,
                                nargs="?",
                                help="path to input file")

        targets = [name for name in self.available_targets]
        arg_parser.add_argument("-t",
                                "--target",
                                type=str,
                                required=False,
                                choices=targets,
                                help="target",
                                default="xdsl")

        frontends = [name for name in self.available_frontends]
        arg_parser.add_argument(
            "-f",
            "--frontend",
            type=str,
            required=False,
            choices=frontends,
            help="Frontend to be used for the input. If not set, "
            "the xdsl frontend or the one for the file extension "
            "is used.")

        arg_parser.add_argument("--disable-verify",
                                default=False,
                                action='store_true')
        arg_parser.add_argument("-o",
                                "--output-file",
                                type=str,
                                required=False,
                                help="path to output file")

        pass_names = ",".join([name for name in self.available_passes])
        arg_parser.add_argument("-p",
                                "--passes",
                                required=False,
                                help="Delimited list of passes."
                                f" Available passes are: {pass_names}",
                                type=str,
                                default="")

        arg_parser.add_argument("--print-between-passes",
                                default=False,
                                action='store_true',
                                help="Print the IR between each pass")

        arg_parser.add_argument("--verify-diagnostics",
                                default=False,
                                action='store_true',
                                help="Prints the content of a triggered "
                                "exception and exits with code 0")

        arg_parser.add_argument(
            "--use-mlir-bindings",
            default=False,
            action='store_true',
            help="Use the MLIR bindings for printing MLIR. "
            "This requires the MLIR Python bindings to be installed.")

        arg_parser.add_argument(
            "--allow-unregistered-ops",
            default=False,
            action='store_true',
            help="Allow the parsing of unregistered operations.")

        arg_parser.add_argument(
            "--generate-coverage",
            default=False,
            action='store_true',
            help="Generate the xDSL code coverage for this run.")

        arg_parser.add_argument(
            "--exec-root",
            type=str,
            default=False,
            required=False,
            help="Defines the directory xdsl-opt will be run in."
            "This flag only takes effect if `--generate-config` was specified."
        )

    def register_all_dialects(self):
        """
        Register all dialects that can be used.

        Add other/additional dialects by overloading this function.
        """
        self.ctx.register_dialect(Builtin)
        self.ctx.register_dialect(Func)
        self.ctx.register_dialect(Arith)
        self.ctx.register_dialect(MemRef)
        self.ctx.register_dialect(Affine)
        self.ctx.register_dialect(Scf)
        self.ctx.register_dialect(Cf)
        self.ctx.register_dialect(CMath)
        self.ctx.register_dialect(IRDL)
        self.ctx.register_dialect(LLVM)
        self.ctx.register_dialect(Vector)

    def register_all_frontends(self):
        """
        Register all frontends that can be used.

        Add other/additional frontends by overloading this function.
        """

        def parse_xdsl(io: IOBase):
            return XDSLParser(self.ctx, io.read(), self.get_input_name(),
                              self.args.allow_unregistered_ops).parse_module()

        def parse_mlir(io: IOBase):
            return MLIRParser(self.ctx, io.read(), self.get_input_name(),
                              self.args.allow_unregistered_ops).parse_module()

        self.available_frontends['xdsl'] = parse_xdsl
        self.available_frontends['mlir'] = parse_mlir

    def register_all_passes(self):
        """
        Register all passes that can be used.

        Add other/additional passes by overloading this function.
        """
        pass

    def register_all_targets(self):
        """
        Register all targets that can be used.

        Add other/additional targets by overloading this function.
        """

        def _output_xdsl(prog: ModuleOp, output: IOBase):
            printer = Printer(stream=output)
            printer.print_op(prog)

        def _output_mlir(prog: ModuleOp, output: IOBase):
            if self.args.use_mlir_bindings:
                from xdsl.mlir_converter import MLIRConverter
                converter = MLIRConverter(self.ctx)
                mlir_module = converter.convert_module(prog)
                print(mlir_module, file=output)
            else:
                printer = Printer(stream=output, target=Printer.Target.MLIR)
                printer.print_op(prog)

        def _output_irdl(prog: ModuleOp, output: IOBase):
            irdl_to_mlir = IRDLPrinter(stream=output)
            irdl_to_mlir.print_module(prog)

        self.available_targets['xdsl'] = _output_xdsl
        self.available_targets['irdl'] = _output_irdl
        self.available_targets['mlir'] = _output_mlir

    def setup_pipeline(self):
        """
        Creates a pipeline that consists of all the passes specified.

        Failes, if not all passes are registered.
        """
        pipeline = [
            str(item) for item in self.args.passes.split(',') if len(item) > 0
        ]

        for p in pipeline:
            if p not in self.available_passes:
                raise Exception(f"Unrecognized pass: {p}")

        self.pipeline = [(p, lambda op, p=p: self.available_passes[p]
                          (self.ctx, op)) for p in pipeline]

    def parse_input(self) -> ModuleOp:
        """
        Parse the input file by invoking the parser specified by the `parser`
        argument. If not set, the parser registered for this file extension
        is used.
        """
        if self.args.input_file is None:
            f = sys.stdin
            file_extension = 'xdsl'
        else:
            f = open(self.args.input_file)
            _, file_extension = os.path.splitext(self.args.input_file)
            file_extension = file_extension.replace(".", "")

        if self.args.frontend:
            file_extension = self.args.frontend

        if file_extension not in self.available_frontends:
            raise Exception(f"Unrecognized file extension '{file_extension}'")

        return self.available_frontends[file_extension](f)

    def apply_passes(self, prog: ModuleOp):
        """Apply passes in order."""
        assert isinstance(prog, ModuleOp)
        if not self.args.disable_verify:
            prog.verify()
        for pass_name, p in self.pipeline:
            p(prog)
            assert isinstance(prog, ModuleOp)
            if not self.args.disable_verify:
                prog.verify()
            if self.args.print_between_passes:
                print(f"IR after {pass_name}:")
                printer = Printer(stream=sys.stdout)
                printer.print_op(prog)
                print("\n\n")

    def output_resulting_program(self, prog: ModuleOp) -> str:
        """Get the resulting program."""
        output = StringIO()
        if self.args.target not in self.available_targets:
            raise Exception(f"Unknown target {self.args.target}")

        self.available_targets[self.args.target](prog, output)
        return output.getvalue()

    def print_to_output_stream(self, contents: str):
        """Print the contents in the expected stream."""
        if self.args.output_file is None:
            print(contents)
        else:
            output_stream = open(self.args.output_file, 'w')
            output_stream.write(contents)

    def get_input_name(self):
        return self.args.input_file or 'stdin'



from __future__ import annotations

from typing import Annotated, TypeVar, Optional, List, TypeAlias

from xdsl.dialects.builtin import (IntegerAttr, IndexType, ArrayAttr,
                                   IntegerType, FlatSymbolRefAttr, StringAttr,
                                   DenseIntOrFPElementsAttr)
from xdsl.ir import (MLIRType, Operation, SSAValue, ParametrizedAttribute,
                     Dialect, OpResult)
from xdsl.irdl import (OptOpAttr, irdl_attr_definition, irdl_op_definition,
                       builder, ParameterDef, Generic, Attribute, AnyAttr,
                       Operand, VarOperand, AttrSizedOperandSegments, OpAttr)

_MemRefTypeElement = TypeVar("_MemRefTypeElement", bound=Attribute)

AnyIntegerAttr: TypeAlias = IntegerAttr[IntegerType | IndexType]


@irdl_attr_definition
class MemRefType(Generic[_MemRefTypeElement], ParametrizedAttribute, MLIRType):
    name = "memref"

    shape: ParameterDef[ArrayAttr[AnyIntegerAttr]]
    element_type: ParameterDef[_MemRefTypeElement]

    def get_num_dims(self) -> int:
        return len(self.shape.data)

    def get_shape(self) -> List[int]:
        return [i.value.data for i in self.shape.data]

    @staticmethod
    @builder
    def from_element_type_and_shape(
            referenced_type: _MemRefTypeElement,
            shape: List[int | AnyIntegerAttr]
    ) -> MemRefType[_MemRefTypeElement]:
        return MemRefType([
            ArrayAttr[AnyIntegerAttr].from_list(
                [IntegerAttr.build(d) for d in shape]), referenced_type
        ])

    @staticmethod
    @builder
    def from_params(
        referenced_type: _MemRefTypeElement,
        shape: ArrayAttr[AnyIntegerAttr] = ArrayAttr.from_list(
            [IntegerAttr.from_int_and_width(1, 64)])
    ) -> MemRefType[_MemRefTypeElement]:
        return MemRefType([shape, referenced_type])


_UnrankedMemrefTypeElems = TypeVar("_UnrankedMemrefTypeElems",
                                   bound=Attribute,
                                   covariant=True)


@irdl_attr_definition
class UnrankedMemrefType(Generic[_UnrankedMemrefTypeElems],
                         ParametrizedAttribute, MLIRType):
    name = "unranked_memref"

    element_type: ParameterDef[_UnrankedMemrefTypeElems]

    @staticmethod
    @builder
    def from_type(
        referenced_type: _UnrankedMemrefTypeElems
    ) -> UnrankedMemrefType[_UnrankedMemrefTypeElems]:
        return UnrankedMemrefType([referenced_type])


AnyUnrankedMemrefType: TypeAlias = UnrankedMemrefType[Attribute]


@irdl_op_definition
class Load(Operation):
    name = "memref.load"
    memref: Annotated[Operand, MemRefType]
    indices: Annotated[VarOperand, IndexType]
    res: Annotated[OpResult, AnyAttr()]

    # TODO varargs for indexing, which must match the memref dimensions
    # Problem: memref dimensions require variadic type parameters,
    # which is subject to change

    def verify_(self):
        if self.memref.typ.element_type != self.res.typ:
            raise Exception(
                "expected return type to match the MemRef element type")

        if self.memref.typ.get_num_dims() != len(self.indices):
            raise Exception("expected an index for each dimension")

    @staticmethod
    def get(ref: SSAValue | Operation,
            indices: List[SSAValue | Operation]) -> Load:
        return Load.build(operands=[ref, indices],
                          result_types=[SSAValue.get(ref).typ.element_type])


@irdl_op_definition
class Store(Operation):
    name = "memref.store"
    value: Annotated[Operand, AnyAttr()]
    memref: Annotated[Operand, MemRefType]
    indices: Annotated[VarOperand, IndexType]

    def verify_(self):
        if self.memref.typ.element_type != self.value.typ:
            raise Exception(
                "Expected value type to match the MemRef element type")

        if self.memref.typ.get_num_dims() != len(self.indices):
            raise Exception("Expected an index for each dimension")

    @staticmethod
    def get(value: Operation | SSAValue, ref: Operation | SSAValue,
            indices: List[Operation | SSAValue]) -> Store:
        return Store.build(operands=[value, ref, indices])


@irdl_op_definition
class Alloc(Operation):
    name = "memref.alloc"

    dynamic_sizes: Annotated[VarOperand, IndexType]
    symbol_operands: Annotated[VarOperand, IndexType]

    memref: Annotated[OpResult, MemRefType]

    # TODO how to constraint the IntegerAttr type?
    alignment: OpAttr[AnyIntegerAttr]

    irdl_options = [AttrSizedOperandSegments()]

    @staticmethod
    def get(return_type: Attribute,
            alignment: int,
            shape: Optional[List[int | AnyIntegerAttr]] = None) -> Alloc:
        if shape is None:
            shape = [1]
        return Alloc.build(operands=[[], []],
                           result_types=[
                               MemRefType.from_element_type_and_shape(
                                   return_type, shape)
                           ],
                           attributes={
                               "alignment":
                               IntegerAttr.from_int_and_width(alignment, 64)
                           })


@irdl_op_definition
class Alloca(Operation):
    name = "memref.alloca"

    dynamic_sizes: Annotated[VarOperand, IndexType]
    symbol_operands: Annotated[VarOperand, IndexType]

    memref: Annotated[OpResult, MemRefType]

    # TODO how to constraint the IntegerAttr type?
    alignment: OpAttr[AnyIntegerAttr]

    irdl_options = [AttrSizedOperandSegments()]

    @staticmethod
    def get(return_type: Attribute,
            alignment: int,
            shape: Optional[List[int | AnyIntegerAttr]] = None) -> Alloca:
        if shape is None:
            shape = [1]
        return Alloca.build(operands=[[], []],
                            result_types=[
                                MemRefType.from_element_type_and_shape(
                                    return_type, shape)
                            ],
                            attributes={
                                "alignment":
                                IntegerAttr.from_int_and_width(alignment, 64)
                            })


@irdl_op_definition
class Dealloc(Operation):
    name = "memref.dealloc"
    memref: Annotated[Operand, MemRefType]

    @staticmethod
    def get(operand: Operation | SSAValue) -> Dealloc:
        return Dealloc.build(operands=[operand])


@irdl_op_definition
class Dealloca(Operation):
    name = "memref.dealloca"
    memref: Annotated[Operand, MemRefType]

    @staticmethod
    def get(operand: Operation | SSAValue) -> Dealloca:
        return Dealloca.build(operands=[operand])


@irdl_op_definition
class GetGlobal(Operation):
    name = "memref.get_global"
    memref: Annotated[OpResult, MemRefType]

    def verify_(self) -> None:
        if 'name' not in self.attributes:
            raise Exception("GetGlobal requires a 'name' attribute")

        if not isinstance(self.attributes['name'], FlatSymbolRefAttr):
            raise Exception(
                "expected 'name' attribute to be a FlatSymbolRefAttr")

    @staticmethod
    def get(name: str, return_type: Attribute) -> GetGlobal:
        return GetGlobal.build(
            result_types=[return_type],
            attributes={"name": FlatSymbolRefAttr.build(name)})

    # TODO how to verify the types, as the global might be defined in another
    # compilation unit


@irdl_op_definition
class Global(Operation):
    name = "memref.global"

    sym_name: OpAttr[StringAttr]
    sym_visibility: OpAttr[StringAttr]
    type: OpAttr[Attribute]
    initial_value: OptOpAttr[Attribute]

    def verify_(self) -> None:
        if not isinstance(self.type, MemRefType):
            raise Exception("Global expects a MemRefType")

        if self.initial_value and not isinstance(self.initial_value,
                                                 DenseIntOrFPElementsAttr):
            raise Exception(
                "Global expects an initial value with type DenseIntOrFPElementsAttr"
            )

    @staticmethod
    def get(sym_name: str | StringAttr,
            typ: Attribute,
            initial_value: Optional[Attribute],
            sym_visibility: str = "private") -> Global:
        return Global.build(
            attributes={
                "sym_name": sym_name,
                "type": typ,
                "initial_value": initial_value,
                "sym_visibility": sym_visibility
            })


MemRef = Dialect([Load, Store, Alloc, Alloca, Dealloc, GetGlobal, Global],
                 [MemRefType, UnrankedMemrefType])



import ast

from dataclasses import dataclass, field
from io import StringIO
from typing import Any, Dict, List
from xdsl.dialects.builtin import ModuleOp
from xdsl.frontend.code_generation import CodeGeneration
from xdsl.frontend.exception import FrontendProgramException
from xdsl.frontend.type_conversion import TypeConverter
from xdsl.printer import Printer


@dataclass
class FrontendProgram:
    """
    Class to store the Python AST of a program written in the frontend. This
    program can be compiled and translated to xDSL or MLIR.
    """

    stmts: List[ast.stmt] | None = field(default=None)
    """AST nodes stored for compilation to xDSL."""

    globals: Dict[str, Any] | None = field(default=None)
    """Global information for this program, including all the imports."""

    xdsl_program: ModuleOp | None = field(default=None)
    """Generated xDSL program when AST is compiled."""

    def _check_can_compile(self):
        if self.stmts is None or self.globals is None:
            msg = \
                """
Cannot compile program without the code context. Try to use:
    p = FrontendProgram()
    with CodeContext(p):
        # Your code here."""
            raise FrontendProgramException(msg)

    def compile(self, desymref=True) -> None:
        """Generates xDSL from the source program."""

        # Both statements and globals msut be initialized from within the
        # `CodeContext`.
        self._check_can_compile()
        assert self.globals is not None
        assert self.stmts is not None

        type_converter = TypeConverter(self.globals)
        self.xdsl_program = CodeGeneration.run_with_type_converter(
            type_converter, self.stmts)
        self.xdsl_program.verify()

        # Optionally run desymrefication pass to produce actual SSA.
        if desymref:
            self.desymref()

    def desymref(self) -> None:
        """Desymrefy the generated xDSL."""

        # TODO: Land desymref in the next patch.
        raise FrontendProgramException(
            "Running desymref pass is not supported.")

        self.xdsl_program.verify()

    def _check_can_print(self):
        if self.xdsl_program is None:
            msg = \
                """
Cannot print the program IR without compiling it first. Make sure to use:
    p = FrontendProgram()
    with CodeContext(p):
        # Your code here.
    p.compile()"""
            raise FrontendProgramException(msg)

    def _print(self, target) -> str:
        self._check_can_print()
        assert self.xdsl_program is not None

        file = StringIO("")
        printer = Printer(stream=file, target=target)
        printer.print_op(self.xdsl_program)
        return file.getvalue().strip()

    def mlir(self) -> str:
        return self._print(Printer.Target.MLIR)

    def xdsl(self) -> str:
        return self._print(Printer.Target.XDSL)



import ast
import xdsl.dialects.builtin as builtin
import xdsl.dialects.func as func
import xdsl.frontend.symref as symref

from dataclasses import dataclass, field
from typing import Any, Dict, List
from xdsl.frontend.exception import CodeGenerationException
from xdsl.frontend.op_inserter import OpInserter
from xdsl.frontend.type_conversion import TypeConverter
from xdsl.ir import Attribute, Block, Region


@dataclass
class CodeGeneration:

    @staticmethod
    def run_with_type_converter(type_converter: TypeConverter,
                                stmts: List[ast.stmt]) -> builtin.ModuleOp:
        """Generates xDSL code and returns it encapsulated into a single module."""
        module = builtin.ModuleOp.from_region_or_ops([])
        visitor = CodegGenerationVisitor(type_converter, module)
        for stmt in stmts:
            visitor.visit(stmt)
        return module


@dataclass
class CodegGenerationVisitor(ast.NodeVisitor):
    """Visitor that generates xDSL from the Python AST."""

    type_converter: TypeConverter = field(init=False)
    """Used for type conversion during code generation."""

    globals: Dict[str, Any] = field(init=False)
    """
    Imports and other global information from the module, useful for looking
    up classes, etc.
    """

    inserter: OpInserter = field(init=False)
    """Used for inserting newly generated operations to the right block."""

    symbol_table: Dict[str, Attribute] | None = field(default=None)
    """
    Maps local variable names to their xDSL types. A single dictionary is sufficient
    because inner functions and global variables are not allowed (yet).
    """

    def __init__(self, type_converter: TypeConverter,
                 module: builtin.ModuleOp) -> None:
        self.type_converter = type_converter
        self.globals = type_converter.globals

        assert len(module.body.blocks) == 1
        self.inserter = OpInserter(module.body.blocks[0])

    def visit(self, node: ast.AST) -> None:
        super().visit(node)

    def generic_visit(self, node: ast.AST) -> None:
        raise CodeGenerationException(
            node.lineno, node.col_offset,
            f"Unsupported Python AST node {str(node)}")

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        # TODO: Implement assignemnt in the next patch.
        pass

    def visit_Assign(self, node: ast.Assign) -> None:
        # TODO: Implement assignemnt in the next patch.
        pass

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:

        # Set the symbol table.
        assert self.symbol_table is None
        self.symbol_table = dict()

        # Then, convert types in the function signature.
        argument_types: List[Attribute] = []
        for i, arg in enumerate(node.args.args):
            if arg.annotation is None:
                raise CodeGenerationException(arg.lineno, arg.col_offset, f"")
            xdsl_type = self.type_converter.convert_type_hint(arg.annotation)
            argument_types.append(xdsl_type)

        return_types: List[Attribute] = []
        if node.returns is not None:
            xdsl_type = self.type_converter.convert_type_hint(node.returns)
            return_types.append(xdsl_type)

        # Create a function operation.
        entry_block = Block()
        body_region = Region.from_block_list([entry_block])
        func_op = func.FuncOp.from_region(node.name, argument_types,
                                          return_types, body_region)
        self.inserter.insert_op(func_op)
        self.inserter.set_insertion_point_from_block(entry_block)

        # All arguments are declared using symref.
        for i, arg in enumerate(node.args.args):
            symbol_name = str(arg.arg)
            block_arg = entry_block.insert_arg(argument_types[i], i)
            self.symbol_table[symbol_name] = argument_types[i]
            entry_block.add_op(symref.Declare.get(symbol_name))
            entry_block.add_op(symref.Update.get(symbol_name, block_arg))

        # Parse function body.
        for stmt in node.body:
            self.visit(stmt)

        # When function definition is processed, reset the symbol table and set
        # the insertion point.
        self.symbol_table = None
        parent_op = func_op.parent_op()
        assert parent_op is not None
        self.inserter.set_insertion_point_from_op(parent_op)

    def visit_Pass(self, node: ast.Pass) -> None:
        parent_op = self.inserter.insertion_point.parent_op()

        # We might have to add an explicit return statement in this case. Make sure to
        # check the type signature.
        if parent_op is not None and isinstance(parent_op, func.FuncOp):
            return_types = parent_op.function_type.outputs.data

            if len(return_types) != 0:
                function_name = parent_op.attributes["sym_name"].data
                raise CodeGenerationException(
                    node.lineno, node.col_offset,
                    f"Expected '{function_name}' to return a type.")
            self.inserter.insert_op(func.Return.get())



import pytest

from xdsl.ir import MLContext, ParametrizedAttribute
from xdsl.irdl import irdl_op_definition, irdl_attr_definition, Operation


@irdl_op_definition
class DummyOp(Operation):
    name = "dummy"


@irdl_op_definition
class DummyOp2(Operation):
    name = "dummy2"


@irdl_attr_definition
class DummyAttr(ParametrizedAttribute):
    name = "dummy_attr"


@irdl_attr_definition
class DummyAttr2(ParametrizedAttribute):
    name = "dummy_attr2"


def test_registration_exceptions():
    ctx = MLContext()
    ctx.register_op(DummyOp)
    with pytest.raises(Exception):
        ctx.register_op(DummyOp)

    ctx.register_attr(DummyAttr)
    with pytest.raises(Exception):
        ctx.register_attr(DummyAttr)


def test_get_exceptions():
    ctx = MLContext()
    ctx.register_op(DummyOp)
    ctx.register_attr(DummyAttr)

    _ = ctx.get_op("dummy")
    with pytest.raises(Exception):
        _ = ctx.get_op("dummy2")

    _ = ctx.get_optional_attr("dummy_attr")
    assert ctx.get_optional_attr("dummy_attr2") is None

    with pytest.raises(Exception):
        _ = ctx.get_attr("dummy_attr2")



from __future__ import annotations
from dataclasses import dataclass, field

from xdsl.ir import IRNode, Block, Operation, Region
from xdsl.utils.exceptions import DiagnosticException
from io import StringIO


@dataclass
class Diagnostic:
    op_messages: dict[Operation, list[str]] = field(default_factory=dict)

    def add_message(self, op: Operation, message: str) -> None:
        """Add a message to an operation."""
        self.op_messages.setdefault(op, []).append(message)

    def raise_exception(
            self,
            message: str,
            ir: IRNode,
            exception_type: type[Exception] = DiagnosticException) -> None:
        """Raise an exception, that will also print all messages in the IR."""
        from xdsl.printer import Printer
        f = StringIO()
        p = Printer(stream=f, diagnostic=self)
        toplevel = ir.get_toplevel_object()
        if isinstance(toplevel, Operation):
            p.print_op(toplevel)
        elif isinstance(toplevel, Block):
            p._print_named_block(toplevel)  # type: ignore
        elif isinstance(toplevel, Region):
            p._print_region(toplevel)  # type: ignore
        else:
            assert "xDSL internal error: get_toplevel_object returned unknown construct"

        raise exception_type(message + "\n\n" + f.getvalue())



from __future__ import annotations

import json
from dataclasses import dataclass, field
from enum import Enum
from frozenlist import FrozenList
from typing import Iterable, TypeVar, Any, Dict, Optional, List, cast

from xdsl.dialects.memref import AnyUnrankedMemrefType, MemRefType, UnrankedMemrefType
from xdsl.ir import (BlockArgument, MLIRType, SSAValue, Block, Callable,
                     Attribute, Region, Operation, Data, ParametrizedAttribute)
from xdsl.utils.diagnostic import Diagnostic
from xdsl.dialects.builtin import (
    AnyIntegerAttr, AnyFloatAttr, AnyUnrankedTensorType, AnyVectorType,
    DenseIntOrFPElementsAttr, Float16Type, Float32Type, Float64Type, FloatAttr,
    IndexType, IntegerType, NoneAttr, OpaqueAttr, Signedness, StringAttr,
    FlatSymbolRefAttr, IntegerAttr, ArrayAttr, IntAttr, TensorType, UnitAttr,
    FunctionType, UnrankedTensorType, UnregisteredOp, VectorType,
    DictionaryAttr)

indentNumSpaces = 2


@dataclass(eq=False, repr=False)
class Printer:

    class Target(Enum):
        XDSL = 1
        MLIR = 2

    stream: Optional[Any] = field(default=None)
    print_generic_format: bool = field(default=False)
    diagnostic: Diagnostic = field(default_factory=Diagnostic)
    target: Target = field(default=Target.XDSL)

    _indent: int = field(default=0, init=False)
    _ssa_values: Dict[SSAValue, str] = field(default_factory=dict, init=False)
    _ssa_names: Dict[str, int] = field(default_factory=dict, init=False)
    _block_names: Dict[Block, int] = field(default_factory=dict, init=False)
    _next_valid_name_id: int = field(default=0, init=False)
    _next_valid_block_id: int = field(default=0, init=False)
    _current_line: int = field(default=0, init=False)
    _current_column: int = field(default=0, init=False)
    _next_line_callback: List[Callable[[], None]] = field(default_factory=list,
                                                          init=False)

    def print(self, *argv: Any) -> None:
        for arg in argv:
            if isinstance(arg, str):
                self.print_string(arg)
                continue
            if isinstance(arg, SSAValue):
                self.print_ssa_value(arg)
                continue
            if isinstance(arg, Attribute):
                self.print_attribute(arg)
                continue
            if isinstance(arg, Region):
                self.print_region(arg)
                continue
            if isinstance(arg, Block):
                self.print_block_name(arg)
                continue
            if isinstance(arg, Operation):
                self.print_op(arg)
                continue
            text = str(arg)
            self.print_string(text)

    def print_string(self, text: str) -> None:
        lines = text.split('\n')
        if len(lines) != 1:
            self._current_line += len(lines) - 1
            self._current_column = len(lines[-1])
        else:
            self._current_column += len(lines[-1])
        print(text, end='', file=self.stream)

    def _add_message_on_next_line(self, message: str, begin_pos: int,
                                  end_pos: int):
        """Add a message that will be displayed on the next line."""

        def callback(indent: int = self._indent):
            self._print_message(message, begin_pos, end_pos, indent)

        self._next_line_callback.append(callback)

    def _print_message(self,
                       message: str,
                       begin_pos: int,
                       end_pos: int,
                       indent: int | None = None):
        """
        Print a message.
        This is expected to be called at the beginning of a new line and to create a new
        line at the end.
        [begin_pos, end_pos)
        """
        indent = self._indent if indent is None else indent
        indent_size = indent * indentNumSpaces
        self.print(" " * indent_size)
        message_end_pos = max(map(len, message.split("\n"))) + indent_size + 2
        first_line = (begin_pos - indent_size) * "-" + (
            end_pos - begin_pos) * "^" + (max(message_end_pos, end_pos) -
                                          end_pos) * "-"
        self.print(first_line)
        self._print_new_line(indent=indent, print_message=False)
        for message_line in message.split("\n"):
            self.print("| ")
            self.print(message_line)
            self._print_new_line(indent=indent, print_message=False)
        self.print("-" * (max(message_end_pos, end_pos) - indent_size))
        self._print_new_line(indent=0, print_message=False)

    T = TypeVar('T')
    K = TypeVar('K')
    V = TypeVar('V')

    def print_list(self,
                   elems: Iterable[T],
                   print_fn: Callable[[T], None],
                   delimiter: str = ", ") -> None:
        for i, elem in enumerate(elems):
            if i:
                self.print(delimiter)
            print_fn(elem)

    def print_dictionary(self,
                         elems: dict[K, V],
                         print_key: Callable[[K], None],
                         print_value: Callable[[V], None],
                         delimiter: str = ", ") -> None:
        for i, (key, value) in enumerate(elems.items()):
            if i:
                self.print(delimiter)
            print_key(key)
            self.print("=")
            print_value(value)

    def _print_new_line(self,
                        indent: int | None = None,
                        print_message: bool = True) -> None:
        indent = self._indent if indent is None else indent
        self.print("\n")
        if print_message:
            for callback in self._next_line_callback:
                callback()
            self._next_line_callback = []
        self.print(" " * indent * indentNumSpaces)

    def _get_new_valid_name_id(self) -> str:
        self._next_valid_name_id += 1
        return str(self._next_valid_name_id - 1)

    def _get_new_valid_block_id(self) -> int:
        self._next_valid_block_id += 1
        return self._next_valid_block_id - 1

    def _print_result_value(self, op: Operation, idx: int) -> None:
        val = op.results[idx]
        self.print("%")
        if val in self._ssa_values.keys():
            name = self._ssa_values[val]
        elif val.name:
            curr_ind = self._ssa_names.get(val.name, 0)
            name = val.name + (str(curr_ind) if curr_ind != 0 else "")
            self._ssa_values[val] = name
            self._ssa_names[val.name] = curr_ind + 1
        else:
            name = self._get_new_valid_name_id()
            self._ssa_values[val] = name
        self.print("%s" % name)
        if self.target == self.Target.XDSL:
            self.print(" : ")
            self.print_attribute(val.typ)

    def _print_results(self, op: Operation) -> None:
        results = op.results
        # No results
        if len(results) == 0:
            return

        # One result
        if len(results) == 1:
            self._print_result_value(op, 0)
            self.print(" = ")
            return

        # Multiple results
        self._print_result_value(op, 0)
        for idx in range(1, len(results)):
            self.print(", ")
            self._print_result_value(op, idx)
        self.print(" = ")

    def print_ssa_value(self, value: SSAValue) -> None:
        if ssa_val := self._ssa_values.get(value):
            self.print(f"%{ssa_val}")
        else:
            begin_pos = self._current_column
            self.print("%<UNKNOWN>")
            end_pos = self._current_column
            self._add_message_on_next_line(
                "ERROR: SSAValue is not part of the IR, are you sure all operations "
                "are added before their uses?", begin_pos, end_pos)

    def _print_operand(self, operand: SSAValue) -> None:
        self.print_ssa_value(operand)

        if self.target == self.Target.XDSL:
            self.print(" : ")
            self.print_attribute(operand.typ)

    def _print_ops(self, ops: List[Operation]) -> None:
        self._indent += 1
        for op in ops:
            self._print_new_line()
            self._print_op(op)
        self._indent -= 1
        if len(ops) > 0:
            self._print_new_line()

    def print_block_name(self, block: Block) -> None:
        self.print("^")
        if block not in self._block_names:
            self._block_names[block] = self._get_new_valid_block_id()
        self.print(self._block_names[block])

    def _print_named_block(self, block: Block) -> None:
        self.print_block_name(block)
        if len(block.args) > 0:
            self.print("(")
            self.print_list(block.args, self._print_block_arg)
            self.print(")")
        self.print(":")
        if len(block.ops) > 0:
            self._print_ops(block.ops)
        else:
            self._print_new_line()

    def _print_block_arg(self, arg: BlockArgument) -> None:
        self.print("%")
        name = self._get_new_valid_name_id()
        self._ssa_values[arg] = name
        self.print("%s : " % name)
        self.print_attribute(arg.typ)

    def print_region(self, region: Region) -> None:
        if len(region.blocks) == 0:
            self.print("{}")
            return

        if len(region.blocks) == 1 and len(region.blocks[0].args) == 0:
            self.print("{")
            self._print_ops(region.blocks[0].ops)
            self.print("}")
            return

        self.print("{")
        self._print_new_line()
        for block in region.blocks:
            self._print_named_block(block)
        self.print("}")

    def print_regions(self, regions: List[Region]) -> None:
        if len(regions) == 0:
            return

        if self.target == self.Target.MLIR:
            self.print(" (")
            self.print_list(regions, self.print_region)
            self.print(")")
        else:
            self.print(" ")
            self.print_list(regions, self.print_region, delimiter=" ")

    def _print_operands(self, operands: FrozenList[SSAValue]) -> None:
        if len(operands) == 0:
            self.print("()")
            return

        self.print("(")
        self._print_operand(operands[0])
        for operand in operands[1:]:
            self.print(", ")
            self._print_operand(operand)
        self.print(")")

    def print_paramattr_parameters(
            self,
            params: list[Attribute],
            always_print_brackets: bool = False) -> None:
        if len(params) == 0 and not always_print_brackets:
            return
        self.print("<")
        self.print_list(params, self.print_attribute)
        self.print(">")

    def print_string_literal(self, string: str):
        self.print(json.dumps(string))

    def print_attribute(self, attribute: Attribute) -> None:
        if isinstance(attribute, UnitAttr):
            return

        if isinstance(attribute, IntegerType):
            if self.target == self.Target.XDSL:
                self.print("!")
            if attribute.signedness.data == Signedness.SIGNLESS:
                self.print("i")
            elif attribute.signedness.data == Signedness.SIGNED:
                self.print("si")
            elif attribute.signedness.data == Signedness.UNSIGNED:
                self.print("ui")
            self.print(attribute.width.data)
            return

        if self.target == self.Target.MLIR:
            if isinstance(attribute, Float16Type):
                self.print('f16')
                return
            if isinstance(attribute, Float32Type):
                self.print('f32')
                return
            if isinstance(attribute, Float64Type):
                self.print('f64')
                return

        if isinstance(attribute, StringAttr):
            self.print_string_literal(attribute.data)
            return

        if isinstance(attribute, FlatSymbolRefAttr):
            self.print(f'@{attribute.data.data}')
            return

        if isinstance(attribute, IntegerAttr):
            attribute = cast(AnyIntegerAttr, attribute)

            # boolean shorthands
            if (isinstance((typ := attribute.typ), IntegerType)
                    and typ.width.data == 1):
                self.print("false" if attribute.value.data == 0 else "true")
                return

            width = attribute.parameters[0]
            typ = attribute.parameters[1]
            assert (isinstance(width, IntAttr))
            self.print(width.data)
            self.print(" : ")
            self.print_attribute(typ)
            return

        if isinstance(attribute, FloatAttr):
            value = attribute.value
            typ = cast(FloatAttr[Float16Type | Float32Type | Float64Type],
                       attribute).type
            self.print(value.data)
            self.print(" : ")
            self.print_attribute(typ)
            return

        if isinstance(attribute, ArrayAttr):
            self.print_string("[")
            self.print_list(
                attribute.data,  # type: ignore
                self.print_attribute)
            self.print_string("]")
            return

        if isinstance(attribute, DictionaryAttr):
            self.print_string("{")
            self.print_dictionary(attribute.data, self.print_string_literal,
                                  self.print_attribute)
            self.print_string("}")
            return

        # Function types have an alias in MLIR, but not in xDSL
        if (isinstance(attribute, FunctionType)
                and self.target == self.Target.MLIR):
            self.print("(")
            self.print_list(attribute.inputs.data, self.print_attribute)
            self.print(") -> ")
            outputs = attribute.outputs.data
            if len(outputs) == 1 and not isinstance(outputs[0], FunctionType):
                self.print_attribute(outputs[0])
            else:
                self.print("(")
                self.print_list(outputs, self.print_attribute)
                self.print(")")
            return

        # Dense element types have an alias in MLIR, but not in xDSL
        if (isinstance(attribute, DenseIntOrFPElementsAttr)
                and self.target == self.Target.MLIR):

            def print_dense_list(array: List[AnyIntegerAttr]
                                 | List[AnyFloatAttr], shape: List[int]):

                def print_one_elem(val: Attribute):
                    if isinstance(val, IntegerAttr):
                        self.print(val.value.data)
                    elif isinstance(val, FloatAttr):
                        self.print(val.value.data)
                    else:
                        raise Exception("unexpected attribute type "
                                        "in DenseIntOrFPElementsAttr: "
                                        f"{type(val)}")

                self.print('[')
                if len(shape) > 1:
                    k = len(array) // shape[0]
                    self.print_list(
                        (array[i:i + k] for i in range(0, len(array), k)),
                        lambda subarray: print_dense_list(subarray, shape[1:]))
                else:
                    self.print_list(array, print_one_elem)
                self.print(']')

            self.print("dense<")
            data = attribute.data.data
            shape = attribute.shape if attribute.shape_is_complete else [
                len(data)
            ]
            print_dense_list(data, shape)
            self.print("> : ")
            self.print(attribute.type)
            return

        # vector types have an alias in MLIR, but not in xDSL
        if ((isinstance(attribute, VectorType)
             or isinstance(attribute, TensorType))
                and self.target == self.Target.MLIR):
            attribute = cast(AnyVectorType, attribute)
            self.print(
                "vector<" if isinstance(attribute, VectorType) else "tensor<")
            self.print_list(
                attribute.shape.data, lambda x: self.print(x.value.data)
                if x.value.data != -1 else self.print("?"), "x")
            if len(attribute.shape.data) != 0:
                self.print("x")
            self.print(attribute.element_type)
            self.print(">")
            return

        # Unranked tensors have an alias in MLIR, but not in xDSL
        if (isinstance(attribute, UnrankedTensorType)
                and self.target == self.Target.MLIR):
            attribute = cast(AnyUnrankedTensorType, attribute)
            self.print("tensor<*x")
            self.print(attribute.element_type)
            self.print(">")
            return

        # memref types have an alias in MLIR, but not in xDSL
        if (isinstance(attribute, MemRefType)
                and self.target == self.Target.MLIR):
            attribute = cast(MemRefType[Attribute], attribute)
            self.print("memref<")
            self.print_list(
                attribute.shape.data, lambda x: self.print(x.value.data)
                if x.value.data != -1 else self.print("?"), "x")
            self.print("x", attribute.element_type)
            self.print(">")
            return

        # Unranked tensors have an alias in MLIR, but not in xDSL
        if (isinstance(attribute, UnrankedMemrefType)
                and self.target == self.Target.MLIR):
            attribute = cast(AnyUnrankedMemrefType, attribute)
            self.print("memref<*x")
            self.print(attribute.element_type)
            self.print(">")
            return

        # IndexType has an alias in MLIR, but not in xDSL
        if (isinstance(attribute, IndexType)
                and self.target == self.Target.MLIR):
            self.print("index")
            return

        # opaque attributes have an alias in MLIR, but not in xDSL
        if (isinstance(attribute, OpaqueAttr)
                and self.target == self.Target.MLIR):
            self.print("opaque<", attribute.ident, ", ", attribute.value, ">")
            if not isinstance(attribute.type, NoneAttr):
                self.print(" : ", attribute.type)
            return

        if self.target == self.Target.MLIR:
            # For the MLIR target, we may print differently some attributes
            self.print("!" if isinstance(attribute, MLIRType) else "#")
            self.print(attribute.name)

            if isinstance(attribute, Data):
                self.print("<")
                attribute = cast(Data[Any], attribute)
                attribute.print_parameter(attribute.data, self)
                self.print(">")
                return

            assert isinstance(attribute, ParametrizedAttribute)

            attribute.print_parameters(self)
            return

        if isinstance(attribute, Data):
            self.print(f'!{attribute.name}<')
            attribute = cast(Data[Any], attribute)
            attribute.print_parameter(attribute.data, self)
            self.print(">")
            return

        assert isinstance(
            attribute,
            ParametrizedAttribute), f'{attribute}: {type(attribute)}'

        # Print parametrized attribute with default formatting
        if self.target == self.Target.XDSL and self.print_generic_format:
            self.print(f'!"{attribute.name}"')
            self.print_paramattr_parameters(attribute.parameters,
                                            always_print_brackets=True)
            return

        self.print(f'!{attribute.name}')
        attribute.print_parameters(self)

    def print_successors(self, successors: List[Block]):
        if len(successors) == 0:
            return
        self.print(" (" if self.target == self.Target.XDSL else " [")
        self.print_list(successors, self.print_block_name)
        self.print(")" if self.target == self.Target.XDSL else "]")

    def _print_attr_string(self, attr_tuple: tuple[str, Attribute]) -> None:
        if isinstance(attr_tuple[1], UnitAttr):
            self.print(f"\"{attr_tuple[0]}\"")
        else:
            self.print(f"\"{attr_tuple[0]}\" = ")
            self.print_attribute(attr_tuple[1])

    def _print_op_attributes(self, attributes: Dict[str, Attribute]) -> None:
        if len(attributes) == 0:
            return

        self.print(" ")
        self.print("[" if self.target == Printer.Target.XDSL else "{")

        attribute_list = list(attributes.items())
        self.print_list(attribute_list, self._print_attr_string)

        self.print("]" if self.target == Printer.Target.XDSL else "}")

    def print_op_with_default_format(self, op: Operation) -> None:
        self._print_operands(op.operands)
        self.print_successors(op.successors)

        # We print attributes with the operation in xDSL.
        if self.target == self.Target.XDSL:
            self._print_op_attributes(op.attributes)
            self.print_regions(op.regions)
        else:
            self.print_regions(op.regions)
            self._print_op_attributes(op.attributes)

        # Print the operation type
        if self.target == self.Target.MLIR:
            self.print(" : (")
            self.print_list(op.operands,
                            lambda operand: self.print_attribute(operand.typ))
            self.print(") -> ")
            if len(op.results) == 0:
                self.print("()")
            elif len(op.results) == 1:
                typ = op.results[0].typ
                # Handle ambiguous case
                if isinstance(typ, FunctionType):
                    self.print("(", typ, ")")
                else:
                    self.print(typ)
            else:
                self.print("(")
                self.print_list(
                    op.results,
                    lambda result: self.print_attribute(result.typ))
                self.print(")")

    def _print_op(self, op: Operation) -> None:
        begin_op_pos = self._current_column
        self._print_results(op)
        if isinstance(op, UnregisteredOp):
            self.print(f'"{op.op_name.data}"')
        elif self.print_generic_format or self.target == self.Target.MLIR:
            self.print(f'"{op.name}"')
        else:
            self.print(op.name)
        end_op_pos = self._current_column
        if op in self.diagnostic.op_messages:
            for message in self.diagnostic.op_messages[op]:
                self._add_message_on_next_line(message, begin_op_pos,
                                               end_op_pos)
        if isinstance(op, UnregisteredOp):
            op_name = op.op_name
            del op.attributes["op_name__"]
            self.print_op_with_default_format(op)
            op.attributes["op_name__"] = op_name
        elif self.print_generic_format or self.target == self.Target.MLIR:
            self.print_op_with_default_format(op)
        else:
            op.print(self)

    def print_op(self, op: Operation) -> None:
        self._print_op(op)
        self._print_new_line()



from __future__ import annotations

import ast
import contextlib
import functools
import itertools
import re
import sys
import traceback
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from io import StringIO
from typing import TypeVar, Iterable

from xdsl.utils.exceptions import ParseError, MultipleSpansParseError
from xdsl.dialects.memref import MemRefType, UnrankedMemrefType
from xdsl.dialects.builtin import (
    AnyTensorType, AnyVectorType, Float16Type, Float32Type, Float64Type,
    FloatAttr, FunctionType, IndexType, IntegerType, Signedness, StringAttr,
    IntegerAttr, ArrayAttr, TensorType, UnrankedTensorType, VectorType,
    FlatSymbolRefAttr, DenseIntOrFPElementsAttr, UnregisteredOp, OpaqueAttr,
    NoneAttr, ModuleOp, UnitAttr, i64)
from xdsl.ir import (SSAValue, Block, Callable, Attribute, Operation, Region,
                     BlockArgument, MLContext, ParametrizedAttribute, Data)


@dataclass
class BacktrackingHistory:
    """
    This class holds on to past errors encountered during parsing.

    Given the following error message:
       <unknown>:2:12
         %0 : !invalid = arith.constant() ["value" = 1 : !i32]
               ^^^^^^^
               'invalid' is not a known attribute

       <unknown>:2:7
         %0 : !invalid = arith.constant() ["value" = 1 : !i32]
              ^
              Expected type of value-id here!

    The BacktrackingHistory will contain the outermost error (expected type of value-id here)
    It's parent will be the next error message (not a known attribute).
    Some errors happen in named regions (e.g. "parsing of operation")
    """
    error: ParseError
    parent: BacktrackingHistory | None
    region_name: str | None
    pos: int

    def print_unroll(self, file=sys.stderr):
        if self.parent:
            if self.parent.get_farthest_point() > self.pos:
                self.parent.print_unroll(file)
                self.print(file)
            else:
                self.print(file)
                self.parent.print_unroll(file)

    def print(self, file=sys.stderr):
        print("Parsing of {} failed:".format(self.region_name or "<unknown>"),
              file=file)
        self.error.print_pretty(file=file)

    @functools.cache
    def get_farthest_point(self) -> int:
        """
        Find the farthest this history managed to parse
        """
        if self.parent:
            return max(self.pos, self.parent.get_farthest_point())
        return self.pos

    def iterate(self) -> Iterable[BacktrackingHistory]:
        yield self
        if self.parent:
            yield from self.parent.iterate()

    def __hash__(self):
        return id(self)


@dataclass(frozen=True)
class Span:
    """
    Parts of the input are always passed around as spans, so we know where they originated.
    """

    start: int
    """
    Start of tokens location in source file, global byte offset in file
    """
    end: int
    """
    End of tokens location in source file, global byte offset in file
    """
    input: Input
    """
    The input being operated on
    """

    def __len__(self):
        return self.len

    @property
    def len(self):
        return self.end - self.start

    @property
    def text(self):
        return self.input.content[self.start:self.end]

    def get_line_col(self) -> tuple[int, int]:
        info = self.input.get_lines_containing(self)
        if info is None:
            return -1, -1
        lines, offset_of_first_line, line_no = info
        return line_no, self.start - offset_of_first_line

    def print_with_context(self, msg: str | None = None) -> str:
        """
        returns a string containing lines relevant to the span. The Span's contents
        are highlighted by up-carets beneath them (`^`). The message msg is printed
        along these.
        """
        info = self.input.get_lines_containing(self)
        if info is None:
            return "Unknown location of span {}. Error: ".format(self, msg)
        lines, offset_of_first_line, line_no = info
        # Offset relative to the first line:
        offset = self.start - offset_of_first_line
        remaining_len = max(self.len, 1)
        capture = StringIO()
        print("{}:{}:{}".format(self.input.name, line_no, offset),
              file=capture)
        for line in lines:
            print(line, file=capture)
            if remaining_len < 0:
                continue
            len_on_this_line = min(remaining_len, len(line) - offset)
            remaining_len -= len_on_this_line
            print("{}{}".format(" " * offset, "^" * max(len_on_this_line, 1)),
                  file=capture)
            if msg is not None:
                print("{}{}".format(" " * offset, msg), file=capture)
                msg = None
            offset = 0
        if msg is not None:
            print(msg, file=capture)
        return capture.getvalue()

    def __repr__(self):
        return "{}[{}:{}](text='{}')".format(self.__class__.__name__,
                                             self.start, self.end, self.text)


@dataclass(frozen=True, repr=False)
class StringLiteral(Span):

    def __post_init__(self):
        if len(self) < 2 or self.text[0] != '"' or self.text[-1] != '"':
            raise ParseError(self, "Invalid string literal!")

    @classmethod
    def from_span(cls, span: Span | None) -> StringLiteral | None:
        """
        Convert a normal span into a StringLiteral, to facilitate parsing.

        If argument is None, returns None.
        """
        if span is None:
            return None
        return cls(span.start, span.end, span.input)

    @property
    def string_contents(self):
        # TODO: is this a hack-job?
        return ast.literal_eval(self.text)


@dataclass(frozen=True)
class Input:
    """
    This is a very simple class that is used to keep track of the input.
    """
    content: str = field(repr=False)
    name: str

    @property
    def len(self):
        return len(self.content)

    def __len__(self):
        return self.len

    def get_lines_containing(self,
                             span: Span) -> tuple[list[str], int, int] | None:
        # A pointer to the start of the first line
        start = 0
        line_no = 0
        source = self.content
        while True:
            next_start = source.find('\n', start)
            line_no += 1
            # Handle eof
            if next_start == -1:
                if span.start > len(source):
                    return None
                return [source[start:]], start, line_no
            # As long as the next newline comes before the spans start we can continue
            if next_start < span.start:
                start = next_start + 1
                continue
            # If the whole span is on one line, we are good as well
            if next_start >= span.end:
                return [source[start:next_start]], start, line_no
            while next_start < span.end:
                next_start = source.find('\n', next_start + 1)
            return source[start:next_start].split('\n'), start, line_no

    def at(self, i: int):
        if i >= self.len:
            raise EOFError()
        return self.content[i]


save_t = tuple[int, tuple[str, ...]]


@dataclass
class Tokenizer:
    """
    This class is used to tokenize an Input.

    It provides an interface for backtracking, so you can use:

    with tokenizer.backtracking():
        # Try stuff
        raise ParseError(...)

    and not worry about manually resetting the input position. Backtracking will also
    record errors that happen during backtracking to provide a richer error reporting
    experience.

    It also provides the following methods to inspect the input:

     - next_token(peek) is used to get the next token
        (which just breaks the input as per the rules defined in break_on)
        peek=True doesn't advance the position in the file.
     - next_token_of_pattern(pattern, peek) can be used to get a next token if it
        conforms to a specific pattern. If a literal string is given, it'll check
        if the next characters match. If a regex is given, it will check
        the regex.
     - starts_with(pattern) checks if the input starts with a literal string or
        regex pattern
    """

    input: Input

    pos: int = field(init=False, default=0)
    """
    The position in the input. Points to the first unconsumed character.
    """

    break_on: tuple[str, ...] = ('.', '%', ' ', '(', ')', '[', ']', '{', '}',
                                 '<', '>', ':', '=', '@', '?', '|', '->', '-',
                                 '//', '\n', '\t', '#', '"', "'", ',', '!')
    """
    characters the tokenizer should break on
    """

    history: BacktrackingHistory | None = field(init=False,
                                                default=None,
                                                repr=False)

    last_token: Span | None = field(init=False, default=None, repr=False)

    def __post_init__(self):
        self.last_token = self.next_token(peek=True)

    def save(self) -> save_t:
        """
        Create a checkpoint in the parsing process, useful for backtracking
        """
        return self.pos, self.break_on

    def resume_from(self, save: save_t):
        """
        Resume from a previously saved position.

        Restores the state of the tokenizer to the exact previous position
        """
        self.pos, self.break_on = save

    @contextlib.contextmanager
    def backtracking(self, region_name: str | None = None):
        """
        This context manager can be used to mark backtracking regions.

        When an error is thrown during backtracking, it is recorded and stored together
        with some meta information in the history attribute.

        The backtracker accepts the following exceptions:
         - ParseError: signifies that the region could not be parsed because of (unexpected) syntax errors
         - AssertionError: this error should probably be phased out in favour of the two above
         - EOFError: signals that EOF was reached unexpectedly

        Any other error will be printed to stderr, but backtracking will continue as normal.
        """
        save = self.save()
        starting_position = self.pos
        try:
            yield
            # Clear error history when something doesn't fail
            # This is because we are only interested in the last "cascade" of failures.
            # If a backtracking() completes without failure, something has been parsed (we assume)
            if self.pos > starting_position and self.history is not None:
                self.history = None
        except Exception as ex:
            how_far_we_got = self.pos

            # If we have no error history, start recording!
            if not self.history:
                self.history = self._history_entry_from_exception(
                    ex, region_name, how_far_we_got)

            # If we got further than on previous attempts
            elif how_far_we_got > self.history.get_farthest_point():
                # Throw away history
                self.history = None
                # Generate new history entry,
                self.history = self._history_entry_from_exception(
                    ex, region_name, how_far_we_got)

            # Otherwise, add to exception, if we are in a named region
            elif region_name is not None and how_far_we_got - starting_position > 0:
                self.history = self._history_entry_from_exception(
                    ex, region_name, how_far_we_got)

            self.resume_from(save)

    def _history_entry_from_exception(self, ex: Exception, region: str,
                                      pos: int) -> BacktrackingHistory:
        """
        Given an exception generated inside a backtracking attempt,
        generate a BacktrackingHistory object with the relevant information in it.

        If an unexpected exception type is encountered, print a traceback to stderr
        """
        if isinstance(ex, ParseError):
            return BacktrackingHistory(ex, self.history, region, pos)
        elif isinstance(ex, AssertionError):
            reason = [
                "Generic assertion failure",
                *(reason for reason in ex.args if isinstance(reason, str)),
            ]
            # We assume that assertions fail because of the last read-in token
            if len(reason) == 1:
                tb = StringIO()
                traceback.print_exc(file=tb)
                reason[0] += "\n" + tb.getvalue()

            return BacktrackingHistory(
                ParseError(self.last_token, reason[-1], self.history),
                self.history,
                region,
                pos,
            )
        elif isinstance(ex, EOFError):
            return BacktrackingHistory(
                ParseError(self.last_token, "Encountered EOF", self.history),
                self.history,
                region,
                pos,
            )

        print("Warning: Unexpected error in backtracking:", file=sys.stderr)
        traceback.print_exception(ex, file=sys.stderr)

        return BacktrackingHistory(
            ParseError(self.last_token, "Unexpected exception: {}".format(ex),
                       self.history),
            self.history,
            region,
            pos,
        )

    def next_token(self, peek: bool = False) -> Span:
        """
        Return a Span of the next token, according to the self.break_on rules.

        Can be modified using:
         - peek: don't advance the position, only "peek" at the input

        This will skip over line comments. Meaning it will skip the entire line if it encounters '//'
        """
        i = self.next_pos()
        # Construct the span:
        span = Span(i, self._find_token_end(i), self.input)
        # Advance pointer if not peeking
        if not peek:
            self.pos = span.end

        # Save last token
        self.last_token = span
        return span

    def next_token_of_pattern(self,
                              pattern: re.Pattern | str,
                              peek: bool = False) -> Span | None:
        """
        Return a span that matched the pattern, or nothing. You can choose not to consume the span.
        """
        try:
            start = self.next_pos()
        except EOFError:
            return None

        # Handle search for string literal
        if isinstance(pattern, str):
            if self.starts_with(pattern):
                if not peek:
                    self.pos = start + len(pattern)
                return Span(start, start + len(pattern), self.input)
            return None

        # Handle regex logic
        match = pattern.match(self.input.content, start)
        if match is None:
            return None

        if not peek:
            self.pos = match.end()

        # Save last token
        self.last_token = Span(start, match.end(), self.input)
        return self.last_token

    def consume_peeked(self, peeked_span: Span):
        if peeked_span.start != self.next_pos():
            raise ParseError(peeked_span, "This is not the peeked span!")
        self.pos = peeked_span.end

    def _find_token_end(self, start: int | None = None) -> int:
        """
        Find the point (optionally starting from start) where the token ends
        """
        i = self.next_pos() if start is None else start
        # Search for literal breaks
        for part in self.break_on:
            if self.input.content.startswith(part, i):
                return i + len(part)
        # Otherwise return the start of the next break
        return min(
            filter(
                lambda x: x >= 0,
                (self.input.content.find(part, i) for part in self.break_on),
            ))

    def next_pos(self, i: int | None = None) -> int:
        """
        Find the next starting position (optionally starting from i)

        This will skip line comments!
        """
        i = self.pos if i is None else i
        # Skip whitespaces
        while self.input.at(i).isspace():
            i += 1

        # Skip comments as well
        if self.input.content.startswith("//", i):
            i = self.input.content.find("\n", i) + 1
            return self.next_pos(i)

        return i

    def is_eof(self):
        """
        Check if the end of the input was reached.
        """
        try:
            self.next_pos()
            return False
        except EOFError:
            return True

    @contextlib.contextmanager
    def configured(self, break_on: tuple[str, ...]):
        """
        This is a helper class to allow expressing a temporary change in config, allowing you to write:

        # Parsing double-quoted string now
        string_content = ""
        with tokenizer.configured(break_on=('"', '\\'),):
            # Use tokenizer

        # Now old config is restored automatically

        """
        save = self.save()

        if break_on is not None:
            self.break_on = break_on

        try:
            yield self
        finally:
            self.break_on = save[1]

    def starts_with(self, text: str | re.Pattern) -> bool:
        try:
            start = self.next_pos()
            if isinstance(text, re.Pattern):
                return text.match(self.input.content, start) is None
            return self.input.content.startswith(text, start)
        except EOFError:
            return False


class ParserCommons:
    """
    Collection of common things used in parsing MLIR/IRDL

    """

    integer_literal = re.compile(r"[+-]?([0-9]+|0x[0-9A-Fa-f]+)")
    decimal_literal = re.compile(r"[+-]?([1-9][0-9]*)")
    string_literal = re.compile(r'"(\\[nfvtr"\\]|[^\n\f\v\r"\\])*"')
    float_literal = re.compile(r"[-+]?[0-9]+\.[0-9]*([eE][-+]?[0-9]+)?")
    bare_id = re.compile(r"[A-Za-z_][\w$.]+")
    value_id = re.compile(r"%([0-9]+|([A-Za-z_$.-][\w$.-]*))")
    suffix_id = re.compile(r"([0-9]+|([A-Za-z_$.-][\w$.-]*))")
    """
    suffix-id ::= (digit+ | ((letter|id-punct) (letter|id-punct|digit)*))
    id-punct  ::= [$._-]
    """
    block_id = re.compile(r"\^([0-9]+|([A-Za-z_$.-][\w$.-]*))")
    type_alias = re.compile(r"![A-Za-z_][\w$.]+")
    attribute_alias = re.compile(r"#[A-Za-z_][\w$.]+")
    boolean_literal = re.compile(r"(true|false)")
    # A list of names that are builtin types
    _builtin_type_names = (
        r"[su]?i\d+", r"f\d+", "tensor", "vector", "memref", "complex",
        "opaque", "tuple", "index", "dense"
        # TODO: add all the Float8E4M3FNType, Float8E5M2Type, and BFloat16Type
    )
    builtin_attr_names = ('dense', 'opaque', 'affine_map', 'array',
                          'dense_resource', 'sparse')
    builtin_type = re.compile("(({}))".format(")|(".join(_builtin_type_names)))
    builtin_type_xdsl = re.compile("!(({}))".format(
        ")|(".join(_builtin_type_names)))
    double_colon = re.compile("::")
    comma = re.compile(",")


class BaseParser(ABC):
    """
    Basic recursive descent parser.

    methods marked try_... will attempt to parse, and return None if they failed. If they return None
    they must make sure to restore all state.

    methods marked parse_... will do "greedy" parsing, meaning they consume as much as they can. They will
    also throw an error if the think they should still be parsing. e.g. when parsing a list of numbers
    separated by '::', the following input will trigger an exception:
        1::2::
    Due to the '::' present after the last element. This is useful for parsing lists, as a trailing
    separator is usually considered a syntax error there.

    must_ type parsers are preferred because they are explicit about their failure modes.
    """

    ctx: MLContext
    """xDSL context."""

    ssaValues: dict[str, SSAValue]
    blocks: dict[str, Block]
    forward_block_references: dict[str, list[Span]]
    """
    Blocks we encountered references to before the definition (must be empty after parsing of region completes)
    """

    T_ = TypeVar("T_")
    """
    Type var used for handling function that return single or multiple Spans. Basically the output type
    of all try_parse functions is T_ | None
    """

    allow_unregistered_ops: bool

    def __init__(self,
                 ctx: MLContext,
                 input: str,
                 name: str = '<unknown>',
                 allow_unregistered_ops=False):
        self.tokenizer = Tokenizer(Input(input, name))
        self.ctx = ctx
        self.ssaValues = dict()
        self.blocks = dict()
        self.forward_block_references = dict()
        self.allow_unregistered_ops = allow_unregistered_ops

    def parse_module(self) -> ModuleOp:
        op = self.try_parse_operation()

        if op is None:
            self.raise_error("Could not parse entire input!")

        if isinstance(op, ModuleOp):
            return op
        else:
            self.tokenizer.pos = 0
            self.raise_error("Expected ModuleOp at top level!",
                             self.tokenizer.next_token())

    def get_ssa_val(self, name: Span) -> SSAValue:
        if name.text not in self.ssaValues:
            self.raise_error('SSA Value used before assignment', name)
        return self.ssaValues[name.text]

    def _get_block_from_name(self, block_name: Span) -> Block:
        """
        This function takes a span containing a block id (like `^42`) and returns a block.

        If the block definition was not seen yet, we create a forward declaration.
        """
        name = block_name.text
        if name not in self.blocks:
            self.forward_block_references[name].append(block_name)
            self.blocks[name] = Block()
        return self.blocks[name]

    def parse_block(self) -> Block:
        block_id, args = self._parse_optional_block_label()

        if block_id is None:
            block = Block(self.tokenizer.last_token)
        elif self.forward_block_references.pop(block_id.text,
                                               None) is not None:
            block = self.blocks[block_id.text]
            block.declared_at = block_id
        else:
            if block_id.text in self.blocks:
                raise MultipleSpansParseError(
                    block_id,
                    "Re-declaration of block {}".format(block_id.text),
                    "Originally declared here:",
                    [(self.blocks[block_id.text].declared_at, None)],
                    self.tokenizer.history,
                )
            block = Block(block_id)
            self.blocks[block_id.text] = block

        for i, (name, type) in enumerate(args):
            arg = BlockArgument(type, block, i)
            self.ssaValues[name.text] = arg
            block.args.append(arg)

        while (next_op := self.try_parse_operation()) is not None:
            block.add_op(next_op)

        return block

    def _parse_optional_block_label(
            self) -> tuple[Span | None, list[tuple[Span, Attribute]]]:
        """
        A block label consists of block-id ( `(` block-arg `,` ... `)` )?
        """
        block_id = self.try_parse_block_id()
        arg_list = list()

        if block_id is not None:
            if self.tokenizer.starts_with('('):
                arg_list = self._parse_block_arg_list()

            self.parse_characters(':', 'Block label must end in a `:`!')

        return block_id, arg_list

    def _parse_block_arg_list(self) -> list[tuple[Span, Attribute]]:
        self.parse_characters('(', 'Block arguments must start with `(`')

        args = self.parse_list_of(self.try_parse_value_id_and_type,
                                  "Expected value-id and type here!")

        self.parse_characters(')', 'Expected closing of block arguments!')

        return args

    def try_parse_single_reference(self) -> Span | None:
        with self.tokenizer.backtracking('part of a reference'):
            self.parse_characters('@', "references must start with `@`")
            if (reference := self.try_parse_string_literal()) is not None:
                return reference
            if (reference := self.try_parse_suffix_id()) is not None:
                return reference
            self.raise_error(
                "References must conform to `@` (string-literal | suffix-id)")

    def parse_reference(self) -> list[Span]:
        return self.parse_list_of(
            self.try_parse_single_reference,
            'Expected reference here in the format of `@` (suffix-id | string-literal)',
            ParserCommons.double_colon,
            allow_empty=False)

    def parse_list_of(self,
                      try_parse: Callable[[], T_ | None],
                      error_msg: str,
                      separator_pattern: re.Pattern = ParserCommons.comma,
                      allow_empty: bool = True) -> list[T_]:
        """
        This is a greedy list-parser. It accepts input only in these cases:

         - If the separator isn't encountered, which signals the end of the list
         - If an empty list is allowed, it accepts when the first try_parse fails
         - If an empty separator is given, it instead sees a failed try_parse as the end of the list.

        This means, that the setup will not accept the input and instead raise an error:
            try_parse = parse_integer_literal
            separator = 'x'
            input = 3x4x4xi32
        as it will read [3,4,4], then see another separator, and expects the next try_parse call to succeed
        (which won't as i32 is not a valid integer literal)
        """
        items = list()
        first_item = try_parse()
        if first_item is None:
            if allow_empty:
                return items
            self.raise_error(error_msg)

        items.append(first_item)

        while (match := self.tokenizer.next_token_of_pattern(separator_pattern)
               ) is not None:
            next_item = try_parse()
            if next_item is None:
                # If the separator is emtpy, we are good here
                if separator_pattern.pattern == '':
                    return items
                self.raise_error(error_msg +
                                 ' because was able to match next separator {}'
                                 .format(match.text))
            items.append(next_item)

        return items

    def try_parse_integer_literal(self) -> Span | None:
        return self.tokenizer.next_token_of_pattern(
            ParserCommons.integer_literal)

    def try_parse_decimal_literal(self) -> Span | None:
        return self.tokenizer.next_token_of_pattern(
            ParserCommons.decimal_literal)

    def try_parse_string_literal(self) -> StringLiteral | None:
        return StringLiteral.from_span(
            self.tokenizer.next_token_of_pattern(ParserCommons.string_literal))

    def try_parse_float_literal(self) -> Span | None:
        return self.tokenizer.next_token_of_pattern(
            ParserCommons.float_literal)

    def try_parse_bare_id(self) -> Span | None:
        return self.tokenizer.next_token_of_pattern(ParserCommons.bare_id)

    def try_parse_value_id(self) -> Span | None:
        return self.tokenizer.next_token_of_pattern(ParserCommons.value_id)

    def try_parse_suffix_id(self) -> Span | None:
        return self.tokenizer.next_token_of_pattern(ParserCommons.suffix_id)

    def try_parse_block_id(self) -> Span | None:
        return self.tokenizer.next_token_of_pattern(ParserCommons.block_id)

    def try_parse_boolean_literal(self) -> Span | None:
        return self.tokenizer.next_token_of_pattern(
            ParserCommons.boolean_literal)

    def try_parse_value_id_and_type(self) -> tuple[Span, Attribute] | None:
        with self.tokenizer.backtracking("value id and type"):
            value_id = self.try_parse_value_id()

            if value_id is None:
                self.raise_error("Invalid value-id format!")

            self.parse_characters(':',
                                  'Expected expression (value-id `:` type)')

            type = self.try_parse_type()

            if type is None:
                self.raise_error("Expected type of value-id here!")
            return value_id, type

    def try_parse_type(self) -> Attribute | None:
        if (builtin_type := self.try_parse_builtin_type()) is not None:
            return builtin_type
        if (dialect_type := self.try_parse_dialect_type()) is not None:
            return dialect_type
        return None

    def try_parse_dialect_type_or_attribute(self) -> Attribute | None:
        """
        Parse a type or an attribute.
        """
        kind = self.tokenizer.next_token_of_pattern(re.compile('[!#]'),
                                                    peek=True)

        if kind is None:
            return None

        with self.tokenizer.backtracking("dialect attribute or type"):
            self.tokenizer.consume_peeked(kind)
            if kind.text == '!':
                return self._parse_dialect_type_or_attribute_inner('type')
            else:
                return self._parse_dialect_type_or_attribute_inner('attribute')

    def try_parse_dialect_type(self):
        """
        Parse a dialect type (something prefixed by `!`, defined by a dialect)
        """
        if not self.tokenizer.starts_with('!'):
            return None
        with self.tokenizer.backtracking("dialect type"):
            self.parse_characters('!', "Dialect type must start with a `!`")
            return self._parse_dialect_type_or_attribute_inner('type')

    def try_parse_dialect_attr(self):
        """
        Parse a dialect attribute (something prefixed by `#`, defined by a dialect)
        """
        if not self.tokenizer.starts_with('#'):
            return None
        with self.tokenizer.backtracking("dialect attribute"):
            self.parse_characters('#',
                                  "Dialect attribute must start with a `#`")
            return self._parse_dialect_type_or_attribute_inner('attribute')

    def _parse_dialect_type_or_attribute_inner(self, kind: str):
        type_name = self.tokenizer.next_token_of_pattern(ParserCommons.bare_id)

        if type_name is None:
            self.raise_error("Expected dialect {} name here!".format(kind))

        type_def = self.ctx.get_optional_attr(type_name.text)
        if type_def is None:
            self.raise_error(
                "'{}' is not a know attribute!".format(type_name.text),
                type_name)

        # Pass the task of parsing parameters on to the attribute/type definition
        if issubclass(type_def, ParametrizedAttribute):
            param_list = type_def.parse_parameters(self)
        elif issubclass(type_def, Data):
            self.parse_characters("<", "This attribute must be parametrized!")
            param_list = type_def.parse_parameter(self)
            self.parse_characters(
                ">", "Invalid attribute parametrization, expected `>`!")
        else:
            assert False, "Mathieu said this cannot be."
        return type_def(param_list)

    @abstractmethod
    def try_parse_builtin_type(self) -> Attribute | None:
        """
        parse a builtin-type like i32, index, vector<i32> etc.
        """
        raise NotImplementedError("Subclasses must implement this method!")

    def _parse_builtin_parametrized_type(self,
                                         name: Span) -> ParametrizedAttribute:
        """
        This function is called after we parse the name of a parameterized type such as vector.
        """

        def unimplemented() -> ParametrizedAttribute:
            raise ParseError(name,
                             "Builtin {} not supported yet!".format(name.text))

        builtin_parsers: dict[str, Callable[[], ParametrizedAttribute]] = {
            "vector": self.parse_vector_attrs,
            "memref": self.parse_memref_attrs,
            "tensor": self.parse_tensor_attrs,
            "complex": self.parse_complex_attrs,
            "tuple": unimplemented,
        }

        self.parse_characters("<", "Expected parameter list here!")
        # Get the parser for the type, falling back to the unimplemented warning
        res = builtin_parsers.get(name.text, unimplemented)()
        self.parse_characters(">", "Expected end of parameter list here!")

        return res

    def parse_complex_attrs(self):
        self.raise_error("ComplexType is unimplemented!")

    def parse_memref_attrs(self) -> MemRefType | UnrankedMemrefType:
        dims = self._parse_tensor_or_memref_dims()
        type = self.try_parse_type()
        if dims is None:
            return UnrankedMemrefType.from_type(type)
        return MemRefType.from_element_type_and_shape(type, dims)

    def try_parse_numerical_dims(self,
                                 accept_closing_bracket: bool = False,
                                 lower_bound: int = 1) -> Iterable[int]:
        while (shape_arg :=
               self._try_parse_shape_element(lower_bound)) is not None:
            yield shape_arg
            # Look out for the closing bracket for scalable vector dims
            if accept_closing_bracket and self.tokenizer.starts_with("]"):
                break
            self.parse_characters("x",
                                  "Unexpected end of dimension parameters!")

    def parse_vector_attrs(self) -> AnyVectorType:
        # Also break on 'x' characters as they are separators in dimension parameters
        with self.tokenizer.configured(break_on=self.tokenizer.break_on +
                                       ("x", )):
            shape = list[int](self.try_parse_numerical_dims())
            scaling_shape: list[int] | None = None

            if self.tokenizer.next_token_of_pattern("[") is not None:
                # We now need to parse the scalable dimensions
                scaling_shape = list(self.try_parse_numerical_dims())
                self.parse_characters(
                    "]", "Expected end of scalable vector dimensions here!")
                self.parse_characters(
                    "x", "Expected end of scalable vector dimensions here!")

            if scaling_shape is not None:
                # TODO: handle scaling vectors!
                self.raise_error("Warning: scaling vectors not supported!")
                pass

            type = self.try_parse_type()
            if type is None:
                self.raise_error(
                    "Expected a type at the end of the vector parameters!")

            return VectorType.from_element_type_and_shape(type, shape)

    def _parse_tensor_or_memref_dims(self) -> list[int] | None:
        with self.tokenizer.configured(break_on=self.tokenizer.break_on +
                                       ('x', )):
            # Check for unranked-ness
            if self.tokenizer.next_token_of_pattern('*') is not None:
                # Consume `x`
                self.parse_characters(
                    'x',
                    'Unranked tensors must follow format (`<*x` type `>`)')
            else:
                # Parse rank:
                return list(self.try_parse_numerical_dims(lower_bound=0))

    def parse_tensor_attrs(self) -> AnyTensorType:
        shape = self._parse_tensor_or_memref_dims()
        type = self.try_parse_type()

        if type is None:
            self.raise_error("Expected tensor type here!")

        if self.tokenizer.starts_with(','):
            # TODO: add tensor encoding!
            raise self.raise_error("Parsing tensor encoding is not supported!")

        if shape is None and self.tokenizer.starts_with(','):
            raise self.raise_error("Unranked tensors don't have an encoding!")

        if shape is not None:
            return TensorType.from_type_and_list(type, shape)

        return UnrankedTensorType.from_type(type)

    def _try_parse_shape_element(self, lower_bound: int = 1) -> int | None:
        """
        Parse a shape element, either a decimal integer immediate or a `?`, which evaluates to -1

        immediate cannot be smaller than lower_bound (defaults to 1) (is 0 for tensors and memrefs)
        """
        int_lit = self.try_parse_decimal_literal()

        if int_lit is not None:
            value = int(int_lit.text)
            if value < lower_bound:
                # TODO: this is ugly, it's a raise inside a try_ type function, which should instead just give up
                raise ParseError(
                    int_lit,
                    "Shape element literal cannot be negative or zero!")
            return value

        if self.tokenizer.next_token_of_pattern('?') is not None:
            return -1
        return None

    def _parse_type_params(self) -> list[Attribute]:
        # Consume opening bracket
        self.parse_characters('<', 'Type must be parameterized!')

        params = self.parse_list_of(self.try_parse_type,
                                    'Expected a type here!')

        self.parse_characters('>',
                              'Expected end of type parameterization here!')

        return params

    def expect(self, try_parse: Callable[[], T_ | None],
               error_message: str) -> T_:
        """
        Used to force completion of a try_parse function. Will throw a parse error if it can't
        """
        res = try_parse()
        if res is None:
            self.raise_error(error_message)
        return res

    def raise_error(self, msg: str, at_position: Span | None = None):
        """
        Helper for raising exceptions, provides as much context as possible to them.

        This will, for example, include backtracking errors, if any occurred previously
        """
        if at_position is None:
            at_position = self.tokenizer.next_token(peek=True)

        raise ParseError(at_position, msg, self.tokenizer.history)

    def parse_characters(self, text: str, msg: str) -> Span:
        if (match := self.tokenizer.next_token_of_pattern(text)) is None:
            self.raise_error(msg)
        return match

    @abstractmethod
    def _parse_op_result_list(
            self) -> tuple[list[Span], list[Attribute] | None]:
        raise NotImplementedError()

    def try_parse_operation(self) -> Operation | None:
        with self.tokenizer.backtracking("operation"):
            return self.parse_operation()

    def parse_operation(self) -> Operation:
        result_list, ret_types = self._parse_op_result_list()
        if len(result_list) > 0:
            self.parse_characters(
                '=',
                'Operation definitions expect an `=` after op-result-list!')

        # Check for custom op format
        op_name = self.try_parse_bare_id()
        if op_name is not None:
            op_type = self._get_op_by_name(op_name)
            op = op_type.parse(ret_types, self)
        else:
            # Check for basic op format
            op_name = self.try_parse_string_literal()
            if op_name is None:
                self.raise_error(
                    "Expected an operation name here, either a bare-id, or a string literal!"
                )

            args, successors, attrs, regions, func_type = self._parse_operation_details(
            )

            if ret_types is None:
                assert func_type is not None
                ret_types = func_type.outputs.data

            op_type = self._get_op_by_name(op_name)

            op = op_type.create(
                operands=[self.ssaValues[span.text] for span in args],
                result_types=ret_types,
                attributes=attrs,
                successors=[
                    self.blocks[block_name.text] for block_name in successors
                ],
                regions=regions)

        # Register the result SSA value names in the parser
        for idx, res in enumerate(result_list):
            ssa_val_name = res.text
            if ssa_val_name in self.ssaValues:
                self.raise_error(
                    f"SSA value {ssa_val_name} is already defined", res)
            self.ssaValues[ssa_val_name] = op.results[idx]
            self.ssaValues[ssa_val_name].name = ssa_val_name.lstrip('%')

        return op

    def _get_op_by_name(self, span: Span) -> type[Operation]:
        if isinstance(span, StringLiteral):
            op_name = span.string_contents
        else:
            op_name = span.text

        op_type = self.ctx.get_optional_op(op_name)

        if op_type is not None:
            return op_type

        if self.allow_unregistered_ops:
            return UnregisteredOp.with_name(op_name, self.ctx)

        self.raise_error(f'Unknown operation {op_name}!', span)

    def parse_region(self) -> Region:
        oldSSAVals = self.ssaValues.copy()
        oldBBNames = self.blocks
        oldForwardRefs = self.forward_block_references
        self.blocks = dict()
        self.forward_block_references = defaultdict(list)

        region = Region()

        try:
            self.parse_characters("{", "Regions begin with `{`")
            if self.tokenizer.starts_with("}"):
                region.add_block(Block())
            else:
                # Parse first block
                block = self.parse_block()
                region.add_block(block)

                while self.tokenizer.starts_with("^"):
                    region.add_block(self.parse_block())

            end = self.parse_characters(
                "}", "Reached end of region, expected `}`!")

            if len(self.forward_block_references) > 0:
                raise MultipleSpansParseError(
                    end,
                    "Region ends with missing block declarations for block(s) {}!"
                    .format(', '.join(self.forward_block_references.keys())),
                    'The following block references are dangling:',
                    [(span, "Reference to block \"{}\" without implementation!"
                      .format(span.text)) for span in itertools.chain(
                          *self.forward_block_references.values())],
                    self.tokenizer.history)

            return region
        finally:
            self.ssaValues = oldSSAVals
            self.blocks = oldBBNames
            self.forward_block_references = oldForwardRefs

    def _try_parse_op_name(self) -> Span | None:
        if (str_lit := self.try_parse_string_literal()) is not None:
            return str_lit
        return self.try_parse_bare_id()

    def _parse_attribute_entry(self) -> tuple[Span, Attribute]:
        """
        Parse entry in attribute dict. Of format:

        attribute_entry := (bare-id | string-literal) `=` attribute
        attribute       := dialect-attribute | builtin-attribute
        """
        if (name := self.try_parse_bare_id()) is None:
            name = self.try_parse_string_literal()

        if name is None:
            self.raise_error(
                "Expected bare-id or string-literal here as part of attribute entry!"
            )

        if not self.tokenizer.starts_with('='):
            return name, UnitAttr()

        self.parse_characters(
            "=", "Attribute entries must be of format name `=` attribute!")

        return name, self.parse_attribute()

    @abstractmethod
    def parse_attribute(self) -> Attribute:
        """
        Parse attribute (either builtin or dialect)

        This is different in xDSL and MLIR, so the actuall implementation is provided by the subclass
        """
        raise NotImplementedError()

    def try_parse_attribute(self) -> Attribute | None:
        with self.tokenizer.backtracking("attribute"):
            return self.parse_attribute()

    def _parse_attribute_type(self) -> Attribute:
        """
        Parses `:` type and returns the type
        """
        self.parse_characters(
            ":", "Expected attribute type definition here ( `:` type )")
        return self.expect(
            self.try_parse_type,
            "Expected attribute type definition here ( `:` type )")

    def try_parse_builtin_attr(self) -> Attribute | None:
        """
        Tries to parse a builtin attribute, e.g. a string literal, int, array, etc..
        """
        next_token = self.tokenizer.next_token(peek=True)
        if next_token.text == '"':
            return self.try_parse_builtin_str_attr()
        elif next_token.text == "[":
            return self.try_parse_builtin_arr_attr()
        elif next_token.text == "@":
            return self.try_parse_ref_attr()
        elif next_token.text == '{':
            return self.try_parse_builtin_dict_attr()
        elif next_token.text == '(':
            return self.try_parse_function_type()
        elif next_token.text in ParserCommons.builtin_attr_names:
            return self.try_parse_builtin_named_attr()
        # Order here is important!
        attrs = (self.try_parse_builtin_float_attr,
                 self.try_parse_builtin_int_attr, self.try_parse_builtin_type)

        for attr_parser in attrs:
            if (val := attr_parser()) is not None:
                return val

    def try_parse_builtin_named_attr(self) -> Attribute | None:
        name = self.tokenizer.next_token(peek=True)
        with self.tokenizer.backtracking("Builtin attribute {}".format(
                name.text)):
            self.tokenizer.consume_peeked(name)
            parsers = {
                'dense': self._parse_builtin_dense_attr,
                'opaque': self._parse_builtin_opaque_attr,
            }

            def not_implemented():
                raise NotImplementedError()

            return parsers.get(name.text, not_implemented)()

    def _parse_builtin_dense_attr(self) -> Attribute | None:
        err_msg = "Malformed dense attribute, format must be (`dense<` array-attr `>:` type)"
        self.parse_characters("<", err_msg)
        info = list(self._parse_builtin_dense_attr_args())
        self.parse_characters(">", err_msg)
        self.parse_characters(":", err_msg)
        type = self.expect(self.try_parse_type,
                           "Dense attribute must be typed!")
        return DenseIntOrFPElementsAttr.from_list(type, info)

    def _parse_builtin_opaque_attr(self):
        self.parse_characters("<", "Opaque attribute must be parametrized")
        str_lit_list = self.parse_list_of(self.try_parse_string_literal,
                                          'Expected opaque attr here!')

        if len(str_lit_list) != 2:
            self.raise_error('Opaque expects 2 string literal parameters!')

        self.parse_characters(
            ">", "Unexpected parameters for opaque attr, expected `>`!")

        type = NoneAttr()
        if self.tokenizer.starts_with(':'):
            self.parse_characters(":", "opaque attribute must be typed!")
            type = self.expect(self.try_parse_type,
                               "opaque attribute must be typed!")

        return OpaqueAttr.from_strings(*(span.string_contents
                                         for span in str_lit_list),
                                       type=type)

    def _parse_builtin_dense_attr_args(self) -> Iterable[int | float]:
        """
        dense attribute params must be:

        dense-attr-params           := float-literal | int-literal | list-of-dense-attrs-params
        list-of-dense-attrs-params  := `[` dense-attr-params (`,` dense-attr-params)* `]`
        """

        def try_parse_int_or_float():
            if (literal := self.try_parse_float_literal()) is not None:
                return float(literal.text)
            if (literal := self.try_parse_integer_literal()) is not None:
                return int(literal.text)
            self.raise_error('Expected int or float literal here!')

        if not self.tokenizer.starts_with('['):
            yield try_parse_int_or_float()
            return

        self.parse_characters('[', '')
        while not self.tokenizer.starts_with(']'):
            yield from self._parse_builtin_dense_attr_args()
            if self.tokenizer.next_token_of_pattern(',') is None:
                break
        self.parse_characters(']', '')

    def try_parse_ref_attr(self) -> FlatSymbolRefAttr | None:
        if not self.tokenizer.starts_with("@"):
            return None

        ref = self.parse_reference()

        if len(ref) > 1:
            self.raise_error("Nested refs are not supported yet!", ref[1])

        return FlatSymbolRefAttr.from_str(ref[0].text)

    def try_parse_builtin_int_attr(self) -> IntegerAttr | None:
        bool = self.try_parse_builtin_boolean_attr()
        if bool is not None:
            return bool

        with self.tokenizer.backtracking("built in int attribute"):
            value = self.expect(
                self.try_parse_integer_literal,
                'Integer attribute must start with an integer literal!')
            if self.tokenizer.next_token(peek=True).text != ':':
                return IntegerAttr.from_params(int(value.text), i64)
            type = self._parse_attribute_type()
            return IntegerAttr.from_params(int(value.text), type)

    def try_parse_builtin_float_attr(self) -> FloatAttr | None:
        with self.tokenizer.backtracking("float literal"):
            value = self.expect(
                self.try_parse_float_literal,
                "Float attribute must start with a float literal!",
            )
            # If we don't see a ':' indicating a type signature
            if not self.tokenizer.starts_with(":"):
                return FloatAttr.from_value(float(value.text))

            type = self._parse_attribute_type()
            return FloatAttr.from_value(float(value.text), type)

    def try_parse_builtin_boolean_attr(self) -> IntegerAttr | None:
        span = self.try_parse_boolean_literal()

        if span is None:
            return None

        int_val = ["false", "true"].index(span.text)
        return IntegerAttr.from_params(int_val, IntegerType.from_width(1))

    def try_parse_builtin_str_attr(self):
        if not self.tokenizer.starts_with('"'):
            return None

        with self.tokenizer.backtracking("string literal"):
            literal = self.try_parse_string_literal()
            if literal is None:
                self.raise_error("Invalid string literal")
            return StringAttr.from_str(literal.string_contents)

    def try_parse_builtin_arr_attr(self) -> ArrayAttr | None:
        if not self.tokenizer.starts_with("["):
            return None
        with self.tokenizer.backtracking("array literal"):
            self.parse_characters("[", "Array literals must start with `[`")
            attrs = self.parse_list_of(self.try_parse_attribute,
                                       "Expected array entry!")
            self.parse_characters(
                "]", "Malformed array contents (expected end of array here!")
            return ArrayAttr.from_list(attrs)

    @abstractmethod
    def parse_optional_attr_dict(self) -> dict[str, Attribute]:
        raise NotImplementedError()

    def _attr_dict_from_tuple_list(
            self, tuple_list: list[tuple[Span,
                                         Attribute]]) -> dict[str, Attribute]:
        """
        Convert a list of tuples (Span, Attribute) to a dictionary.

        This function converts the span to a string, trimming quotes from string literals
        """

        def span_to_str(span: Span) -> str:
            if isinstance(span, StringLiteral):
                return span.string_contents
            return span.text

        return dict((span_to_str(span), attr) for span, attr in tuple_list)

    def parse_function_type(self) -> FunctionType:
        """
        Parses function-type:

        viable function types are:
            (i32)   -> ()
            ()      -> (i32, i32)
            (i32, i32) -> ()
            ()      -> i32
        Non-viable types are:
            i32     -> i32
            i32     -> ()

        Uses type-or-type-list-parens internally
        """
        self.parse_characters(
            "(", "First group of function args must start with a `(`")

        args: list[Attribute] = self.parse_list_of(self.try_parse_type,
                                                   "Expected type here!")

        self.parse_characters(
            ")",
            "Malformed function type, expected closing brackets of argument types!"
        )

        self.parse_characters("->", "Malformed function type, expected `->`!")

        return FunctionType.from_lists(args,
                                       self._parse_type_or_type_list_parens())

    def _parse_type_or_type_list_parens(self) -> list[Attribute]:
        """
        Parses type-or-type-list-parens, which is used in function-type.

        type-or-type-list-parens ::= type | type-list-parens
        type-list-parens         ::= `(` `)` | `(` type-list-no-parens `)`
        type-list-no-parens      ::=  type (`,` type)*
        """
        if self.tokenizer.next_token_of_pattern("(") is not None:
            args: list[Attribute] = self.parse_list_of(self.try_parse_type,
                                                       "Expected type here!")
            self.parse_characters(")", "Unclosed function type argument list!")
        else:
            args = [self.try_parse_type()]
            if args[0] is None:
                self.raise_error(
                    "Function type must either be single type or list of types in"
                    " parenthesis!")
        return args

    def try_parse_function_type(self) -> FunctionType | None:
        if not self.tokenizer.starts_with("("):
            return None
        with self.tokenizer.backtracking("function type"):
            return self.parse_function_type()

    def parse_region_list(self) -> list[Region]:
        """
        Parses a sequence of regions for as long as there is a `{` in the input.
        """
        regions = []
        while not self.tokenizer.is_eof() and self.tokenizer.starts_with("{"):
            regions.append(self.parse_region())
        return regions

    def _parse_builtin_type_with_name(self, name: Span):
        """
        Parses one of the builtin types like i42, vector, etc...
        """
        if name.text == "index":
            return IndexType()
        if (re_match := re.match(r"^[su]?i(\d+)$", name.text)) is not None:
            signedness = {
                "s": Signedness.SIGNED,
                "u": Signedness.UNSIGNED,
                "i": Signedness.SIGNLESS,
            }
            return IntegerType.from_width(int(re_match.group(1)),
                                          signedness[name.text[0]])

        if (re_match := re.match(r"^f(\d+)$", name.text)) is not None:
            width = int(re_match.group(1))
            type = {
                16: Float16Type,
                32: Float32Type,
                64: Float64Type
            }.get(width, None)
            if type is None:
                self.raise_error(
                    "Unsupported floating point width: {}".format(width))
            return type()

        return self._parse_builtin_parametrized_type(name)

    @abstractmethod
    def _parse_operation_details(
        self,
    ) -> tuple[list[Span], list[Span], dict[str, Attribute], list[Region],
               FunctionType | None]:
        """
        Must return a tuple consisting of:
            - a list of arguments to the operation
            - a list of successor names
            - the attributes attached to the OP
            - the regions of the op
            - An optional function type. If not supplied, parse_op_result_list must return a second value
              containing the types of the returned SSAValues

        """
        raise NotImplementedError()

    @abstractmethod
    def _parse_op_args_list(self) -> list[Span]:
        raise NotImplementedError()

    # HERE STARTS A SOMEWHAT CURSED COMPATIBILITY LAYER:
    # Since we don't want to rewrite all dialects currently, the new parser needs to expose the same
    # Interface to the dialect definitions (to some extent). Here we implement that interface.

    _OperationType = TypeVar("_OperationType", bound=Operation)

    def parse_op_with_default_format(
        self,
        op_type: type[_OperationType],
        result_types: list[Attribute],
    ) -> _OperationType:
        """
        Compatibility wrapper so the new parser can be passed instead of the old one. Parses everything after the
        operation name.

        This implicitly assumes XDSL format, and will fail on MLIR style operations
        """
        # TODO: remove this function and restructure custom op / irdl parsing
        assert isinstance(self, XDSLParser)
        args, successors, attributes, regions, _ = self._parse_operation_details(
        )

        for x in args:
            if x.text not in self.ssaValues:
                self.raise_error(
                    "Unknown SSAValue name, known SSA Values are: {}".format(
                        ", ".join(self.ssaValues.keys())), x)

        return op_type.create(
            operands=[self.ssaValues[span.text] for span in args],
            result_types=result_types,
            attributes=attributes,
            successors=[
                self._get_block_from_name(span) for span in successors
            ],
            regions=regions)

    def parse_paramattr_parameters(
            self,
            expect_brackets: bool = False,
            skip_white_space: bool = True) -> list[Attribute]:
        opening_brackets = self.tokenizer.next_token_of_pattern('<')
        if expect_brackets and opening_brackets is None:
            self.raise_error("Expected start attribute parameters here (`<`)!")

        res = self.parse_list_of(self.try_parse_attribute,
                                 'Expected another attribute here!')

        if opening_brackets is not None and self.tokenizer.next_token_of_pattern(
                '>') is None:
            self.raise_error(
                "Malformed parameter list, expected either another parameter or `>`!"
            )

        return res

    def parse_char(self, text: str):
        self.parse_characters(text, "Expected '{}' here!".format(text))

    def parse_str_literal(self) -> str:
        return self.expect(self.try_parse_string_literal,
                           'Malformed string literal!').string_contents

    def parse_attribute(self) -> Attribute:
        raise NotImplementedError()

    def parse_op(self) -> Operation:
        return self.parse_operation()

    def parse_int_literal(self) -> int:
        return int(
            self.expect(self.try_parse_integer_literal,
                        'Expected integer literal here').text)

    def try_parse_builtin_dict_attr(self):
        attr_def = self.ctx.get_optional_attr('dictionary')
        if attr_def is None:
            self.raise_error(
                "An attribute named `dictionary` must be available in the "
                "context in order to parse dictionary attributes! Please make "
                "sure the builtin dialect is available, or provide your own "
                "replacement!")
        param = attr_def.parse_parameter(self)
        return attr_def(param)


class MLIRParser(BaseParser):

    def try_parse_builtin_type(self) -> Attribute | None:
        """
        parse a builtin-type like i32, index, vector<i32> etc.
        """
        with self.tokenizer.backtracking("builtin type"):
            name = self.tokenizer.next_token_of_pattern(
                ParserCommons.builtin_type)
            if name is None:
                raise self.raise_error("Expected builtin name!")

            return self._parse_builtin_type_with_name(name)

    def parse_attribute(self) -> Attribute:
        """
        Parse attribute (either builtin or dialect)
        """
        # All dialect attrs must start with '#', so we check for that first (as it's easier)
        if self.tokenizer.starts_with("#"):
            value = self.try_parse_dialect_attr()

            # No value => error
            if value is None:
                self.raise_error(
                    "`#` must be followed by a valid dialect attribute or type!"
                )

            return value

        # If it isn't a dialect attr, parse builtin
        builtin_val = self.try_parse_builtin_attr()

        if builtin_val is None:
            self.raise_error(
                "Unknown attribute (neither builtin nor dialect could be parsed)!"
            )

        return builtin_val

    def _parse_op_result_list(
            self) -> tuple[list[Span], list[Attribute] | None]:
        return (
            self.parse_list_of(self.try_parse_value_id,
                               "Expected op-result here!",
                               allow_empty=True),
            None,
        )

    def parse_optional_attr_dict(self) -> dict[str, Attribute]:
        if not self.tokenizer.starts_with("{"):
            return dict()

        self.parse_characters(
            "{",
            "MLIR Attribute dictionary must be enclosed in curly brackets")

        attrs = []
        if not self.tokenizer.starts_with('}'):
            attrs = self.parse_list_of(self._parse_attribute_entry,
                                       "Expected attribute entry")

        self.parse_characters(
            "}",
            "MLIR Attribute dictionary must be enclosed in curly brackets")

        return self._attr_dict_from_tuple_list(attrs)

    def _parse_operation_details(
        self,
    ) -> tuple[list[Span], list[Span], dict[str, Attribute], list[Region],
               FunctionType | None]:
        args = self._parse_op_args_list()
        succ = self._parse_optional_successor_list()

        regions = []
        if self.tokenizer.starts_with("("):
            self.parse_characters("(", "Expected brackets enclosing regions!")
            regions = self.parse_region_list()
            self.parse_characters(")", "Expected brackets enclosing regions!")

        attrs = self.parse_optional_attr_dict()

        self.parse_characters(
            ":",
            "MLIR Operation definitions must end in a function type signature!"
        )
        func_type = self.parse_function_type()

        return args, succ, attrs, regions, func_type

    def _parse_optional_successor_list(self) -> list[Span]:
        if not self.tokenizer.starts_with("["):
            return []
        self.parse_characters("[",
                              "Successor list is enclosed in square brackets")
        successors = self.parse_list_of(self.try_parse_block_id,
                                        "Expected a block-id",
                                        allow_empty=False)
        self.parse_characters("]",
                              "Successor list is enclosed in square brackets")
        return successors

    def _parse_op_args_list(self) -> list[Span]:
        self.parse_characters(
            "(", "Operation args list must be enclosed by brackets!")
        args = self.parse_list_of(self.try_parse_value_id,
                                  "Expected another bare-id here")
        self.parse_characters(
            ")", "Operation args list must be closed by a closing bracket")
        # TODO: check if type is correct here!
        return args


class XDSLParser(BaseParser):

    def try_parse_builtin_type(self) -> Attribute | None:
        """
        parse a builtin-type like i32, index, vector<i32> etc.
        """
        with self.tokenizer.backtracking("builtin type"):
            name = self.tokenizer.next_token_of_pattern(
                ParserCommons.builtin_type_xdsl)
            if name is None:
                self.raise_error("Expected builtin name!")
            # xDSL builtin types have a '!' prefix, we strip that out here
            name = Span(start=name.start + 1, end=name.end, input=name.input)

            return self._parse_builtin_type_with_name(name)

    def parse_attribute(self) -> Attribute:
        """
        Parse attribute (either builtin or dialect)

        xDSL allows types in places of attributes! That's why we parse types here as well
        """
        value = self.try_parse_builtin_attr()

        # xDSL: Allow both # and ! prefixes, as we allow both types and attrs
        # TODO: phase out use of next_token(peek=True) in favour of starts_with
        if value is None and self.tokenizer.next_token(peek=True).text in "#!":
            # In MLIR # and ! are prefixes for dialect attrs/types, but in xDSL ! is also used for builtin types
            value = self.try_parse_dialect_type_or_attribute()

        if value is None:
            self.raise_error(
                "Unknown attribute (neither builtin nor dialect could be parsed)!"
            )

        return value

    def _parse_op_result_list(
            self) -> tuple[list[Span], list[Attribute] | None]:
        if not self.tokenizer.starts_with("%"):
            return list(), list()
        results = self.parse_list_of(
            self.try_parse_value_id_and_type,
            "Expected (value-id `:` type) here!",
            allow_empty=False,
        )
        # TODO: this is hideous, make it cleaner
        # zip(*results) works, but is barely readable :/
        return [name for name, _ in results], [type for _, type in results]

    def try_parse_builtin_attr(self) -> Attribute:
        """
        Tries to parse a builtin attribute, e.g. a string literal, int, array, etc..

        If the mode is xDSL, it also allows parsing of builtin types
        """
        # In xdsl, two things are different here:
        #  1. types are considered valid attributes
        #  2. all types, builtins included, are prefixed with !
        if self.tokenizer.starts_with("!"):
            return self.try_parse_builtin_type()

        return super().try_parse_builtin_attr()

    def parse_optional_attr_dict(self) -> dict[str, Attribute]:
        if not self.tokenizer.starts_with("["):
            return dict()

        self.parse_characters(
            "[",
            "xDSL Attribute dictionary must be enclosed in square brackets")

        attrs = self.parse_list_of(self._parse_attribute_entry,
                                   "Expected attribute entry")

        self.parse_characters(
            "]",
            "xDSL Attribute dictionary must be enclosed in square brackets")

        return self._attr_dict_from_tuple_list(attrs)

    def _parse_operation_details(
        self,
    ) -> tuple[list[Span], list[Span], dict[str, Attribute], list[Region],
               FunctionType | None]:
        """
        Must return a tuple consisting of:
            - a list of arguments to the operation
            - a list of successor names
            - the attributes attached to the OP
            - the regions of the op
            - An optional function type. If not supplied, parse_op_result_list must return a second value
              containing the types of the returned SSAValues

        """
        args = self._parse_op_args_list()
        succ = self._parse_optional_successor_list()
        attrs = self.parse_optional_attr_dict()
        regions = self.parse_region_list()

        return args, succ, attrs, regions, None

    def _parse_optional_successor_list(self) -> list[Span]:
        if not self.tokenizer.starts_with("("):
            return []
        self.parse_characters("(",
                              "Successor list is enclosed in round brackets")
        successors = self.parse_list_of(self.try_parse_block_id,
                                        "Expected a block-id",
                                        allow_empty=False)
        self.parse_characters(")",
                              "Successor list is enclosed in round brackets")
        return successors

    def _parse_dialect_type_or_attribute_inner(self, kind: str):
        if self.tokenizer.starts_with('"'):
            name = self.try_parse_string_literal()
            if name is None:
                self.raise_error(
                    "Expected string literal for an attribute in generic format here!"
                )
            return self._parse_generic_attribute_args(name)
        return super()._parse_dialect_type_or_attribute_inner(kind)

    def _parse_generic_attribute_args(self, name: StringLiteral):
        attr = self.ctx.get_optional_attr(name.string_contents)
        if attr is None:
            self.raise_error("Unknown attribute name!", name)
        if not issubclass(attr, ParametrizedAttribute):
            self.raise_error("Expected ParametrizedAttribute name here!", name)
        self.parse_characters('<',
                              'Expected generic attribute arguments here!')
        args = self.parse_list_of(self.try_parse_attribute,
                                  'Unexpected end of attribute list!')
        self.parse_characters(
            '>', 'Malformed attribute arguments, reached end of args list!')
        return attr(args)

    def _parse_op_args_list(self) -> list[Span]:
        self.parse_characters(
            "(", "Operation args list must be enclosed by brackets!")
        args = self.parse_list_of(self.try_parse_value_id_and_type,
                                  "Expected another bare-id here")
        self.parse_characters(
            ")", "Operation args list must be closed by a closing bracket")
        # TODO: check if type is correct here!
        return [name for name, _ in args]

    def try_parse_type(self) -> Attribute | None:
        return self.try_parse_attribute()


# COMPAT layer so parser_ng is a drop-in replacement for parser:


class Source(Enum):
    XDSL = 1
    MLIR = 2


def Parser(ctx: MLContext,
           prog: str,
           source: Source = Source.XDSL,
           filename: str = '<unknown>',
           allow_unregistered_ops=False) -> BaseParser:
    selected_parser = {
        Source.XDSL: XDSLParser,
        Source.MLIR: MLIRParser
    }[source]
    return selected_parser(ctx, prog, filename)


setattr(Parser, 'Source', Source)



from __future__ import annotations
from typing import Annotated, List

from xdsl.dialects.builtin import IndexType, VectorType, i1
from xdsl.dialects.memref import MemRefType
from xdsl.ir import Operation, SSAValue, Dialect, OpResult
from xdsl.irdl import AnyAttr, irdl_op_definition, Operand, VarOperand
from xdsl.utils.exceptions import VerifyException


@irdl_op_definition
class Load(Operation):
    name = "vector.load"
    memref: Annotated[Operand, MemRefType]
    indices: Annotated[VarOperand, IndexType]
    res: Annotated[OpResult, VectorType]

    def verify_(self):
        if self.memref.typ.element_type != self.res.typ.element_type:
            raise VerifyException(
                "MemRef element type should match the Vector element type.")

        if self.memref.typ.get_num_dims() != len(self.indices):
            raise VerifyException("Expected an index for each dimension.")

    @staticmethod
    def get(ref: SSAValue | Operation,
            indices: List[SSAValue | Operation]) -> Load:
        return Load.build(operands=[ref, indices],
                          result_types=[
                              VectorType.from_element_type_and_shape(
                                  SSAValue.get(ref).typ.element_type, [1])
                          ])


@irdl_op_definition
class Store(Operation):
    name = "vector.store"
    vector: Annotated[Operand, VectorType]
    memref: Annotated[Operand, MemRefType]
    indices: Annotated[VarOperand, IndexType]

    def verify_(self):
        if self.memref.typ.element_type != self.vector.typ.element_type:
            raise VerifyException(
                "MemRef element type should match the Vector element type.")

        if self.memref.typ.get_num_dims() != len(self.indices):
            raise VerifyException("Expected an index for each dimension.")

    @staticmethod
    def get(vector: Operation | SSAValue, ref: Operation | SSAValue,
            indices: List[Operation | SSAValue]) -> Store:
        return Store.build(operands=[vector, ref, indices])


@irdl_op_definition
class Broadcast(Operation):
    name = "vector.broadcast"
    source: Annotated[Operand, AnyAttr()]
    vector: Annotated[OpResult, VectorType]

    def verify_(self):
        if self.source.typ != self.vector.typ.element_type:
            raise VerifyException(
                "Source operand and result vector must have the same element type."
            )

    @staticmethod
    def get(source: Operation | SSAValue) -> Broadcast:
        return Broadcast.build(operands=[source],
                               result_types=[
                                   VectorType.from_element_type_and_shape(
                                       SSAValue.get(source).typ, [1])
                               ])


@irdl_op_definition
class FMA(Operation):
    name = "vector.fma"
    lhs: Annotated[Operand, VectorType]
    rhs: Annotated[Operand, VectorType]
    acc: Annotated[Operand, VectorType]
    res: Annotated[OpResult, VectorType]

    def verify_(self):
        res_element_type = self.res.typ.element_type
        res_shape = self.res.typ.get_shape()

        if res_element_type != self.lhs.typ.element_type:
            raise VerifyException(
                "Result vector type must match with all source vectors. Found different types for result vector and lhs vector."
            )
        elif res_element_type != self.rhs.typ.element_type:
            raise VerifyException(
                "Result vector type must match with all source vectors. Found different types for result vector and rhs vector."
            )
        elif res_element_type != self.acc.typ.element_type:
            raise VerifyException(
                "Result vector type must match with all source vectors. Found different types for result vector and acc vector."
            )

        if res_shape != self.lhs.typ.get_shape():
            raise VerifyException(
                "Result vector shape must match with all source vector shapes. Found different shapes for result vector and lhs vector."
            )
        elif res_shape != self.rhs.typ.get_shape():
            raise VerifyException(
                "Result vector shape must match with all source vector shapes. Found different shapes for result vector and rhs vector."
            )
        elif res_shape != self.acc.typ.get_shape():
            raise VerifyException(
                "Result vector shape must match with all source vector shapes. Found different shapes for result vector and acc vector."
            )

    @staticmethod
    def get(lhs: Operation | SSAValue, rhs: Operation | SSAValue,
            acc: Operation | SSAValue) -> FMA:
        return FMA.build(operands=[lhs, rhs, acc],
                         result_types=[
                             VectorType.from_element_type_and_shape(
                                 SSAValue.get(lhs).typ.element_type, [1])
                         ])


@irdl_op_definition
class Maskedload(Operation):
    name = "vector.maskedload"
    memref: Annotated[Operand, MemRefType]
    indices: Annotated[VarOperand, IndexType]
    mask: Annotated[Operand, VectorType]
    passthrough: Annotated[Operand, VectorType]
    res: Annotated[OpResult, VectorType]

    def verify_(self):
        memref_element_type = self.memref.typ.element_type

        if memref_element_type != self.res.typ.element_type:
            raise VerifyException(
                "MemRef element type should match the result vector and passthrough vector element type. Found different element types for memref and result."
            )
        elif memref_element_type != self.passthrough.typ.element_type:
            raise VerifyException(
                "MemRef element type should match the result vector and passthrough vector element type. Found different element types for memref and passthrough."
            )

        if len(self.res.typ.get_shape()) != 1:
            raise VerifyException("Expected a rank 1 result vector.")

        if len(self.mask.typ.get_shape()) != 1:
            raise VerifyException("Expected a rank 1 mask vector.")

        if self.mask.typ.element_type != i1:
            raise VerifyException("Expected mask element type to be i1.")

        if self.memref.typ.get_num_dims() != len(self.indices):
            raise VerifyException(
                "Expected an index for each memref dimension.")

    @staticmethod
    def get(memref: SSAValue | Operation, indices: List[SSAValue | Operation],
            mask: SSAValue | Operation,
            passthrough: SSAValue | Operation) -> Maskedload:
        return Maskedload.build(operands=[memref, indices, mask, passthrough],
                                result_types=[
                                    VectorType.from_element_type_and_shape(
                                        SSAValue.get(memref).typ.element_type,
                                        [1])
                                ])


Vector = Dialect([Load, Store, Broadcast, FMA, Maskedload], [])



from io import IOBase
from dataclasses import dataclass

from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import Attribute, Operation
from xdsl.irdl import AttrConstraint
from xdsl.printer import Printer
from xdsl.dialects.irdl import (
    ConstraintVarsOp, DialectOp, ParametersOp, TypeOp, OperandsOp, ResultsOp,
    OperationOp, EqTypeConstraintAttr, AnyTypeConstraintAttr,
    AnyOfTypeConstraintAttr, DynTypeBaseConstraintAttr,
    DynTypeParamsConstraintAttr, TypeParamsConstraintAttr,
    NamedTypeConstraintAttr)


@dataclass(frozen=True, eq=False)
class IRDLPrinter:

    stream: IOBase

    def _print(self, s: str, end: str | None = None):
        print(s, file=self.stream, end=end)

    def print_module(self, module: ModuleOp):
        module.walk(lambda op: self.ensure_op_is_irdl_op(op))
        self._print('module {')
        module.walk(lambda di: IRDLPrinter.print_dialect_definition(self, di)
                    if isinstance(di, DialectOp) else None)
        self._print('}')

    def ensure_op_is_irdl_op(self, op: Operation):
        if not isinstance(
                op, (OperationOp | ResultsOp | OperandsOp, ConstraintVarsOp
                     | TypeOp | ParametersOp | DialectOp | ModuleOp)):
            raise Exception(f"Operation {op.name} is not an operation in IRDL")

    def print_type_definition(self, type: TypeOp):
        self._print(f"    {TypeOp.name} {type.type_name.data} {{")
        type.walk(lambda param: self.print_parameters_definition(param)
                  if isinstance(param, ParametersOp) else None)
        self._print("    }")

    def print_attr_constraint(self, f: AttrConstraint | Attribute):

        if isinstance(f, AnyOfTypeConstraintAttr):
            self._print("AnyOf", end="<")
            for i in range(len(f.params.data) - 1):
                self.print_attr_constraint(f.params.data[i])
                self._print(", ", end="")
            self.print_attr_constraint(f.params.data[len(f.params.data) - 1])
            self._print(">", end="")

        elif isinstance(f, EqTypeConstraintAttr):
            p = Printer(self.stream)
            p.print(f.type)

        elif isinstance(f, AnyTypeConstraintAttr):
            self._print("Any", end="")

        elif isinstance(f, DynTypeBaseConstraintAttr):
            self._print(f.type_name.data, end='')

        elif isinstance(f, DynTypeParamsConstraintAttr):
            self._print(f.type_name.data, end='')
            self._print("<", end="")
            for i in range(len(f.params.data)):
                self.print_attr_constraint(f.params.data[i])
            self._print(">", end="")

        elif isinstance(f, TypeParamsConstraintAttr):
            self._print(f"{f.type_name}")

        elif isinstance(f, NamedTypeConstraintAttr):
            self._print(f.type_name.data, end='')
            self._print(":", end=" ")
            self.print_attr_constraint(f.params_constraints)

        else:
            raise Exception(f"Invalid Constraint: {type(f)} ")

    def print_parameters_definition(self, param_op: ParametersOp):
        self._print(f"      {ParametersOp.name}", end="(")

        for param in param_op.params.data:
            self._print(f"{param.type_name.data}: ", end="")
            self.print_attr_constraint(param.params_constraints)
        self._print(")")

    def print_operation_definition(self, operation: OperationOp):
        self._print(
            f"    {OperationOp.name} {operation.attributes['name'].data} {{")

        # Checking for existence of operands
        operand_list = []
        operation.walk(lambda operand_def: operand_list.append(operand_def)
                       if isinstance(operand_def, OperandsOp) else None)
        if operand_list:
            self.print_operand_definition(operand_list)

        # Checking for existence of results
        result_list = []
        operation.walk(lambda res: result_list.append(res)
                       if isinstance(res, ResultsOp) else None)
        if result_list:
            self.print_result_definition(result_list)

        self._print("    }}")

    def print_result_definition(self, res_list: list[ResultsOp]):
        self._print(f"      {ResultsOp.name}", end="(")

        for i in range(len(res_list)):
            for result in res_list[i].params.data:
                self.print_attr_constraint(result)
                self._print(", ", end='') if i != len(res_list) - 1 else None
        self._print(")")

    def print_operand_definition(self, op_list: list[OperandsOp]):
        self._print(f"      {OperandsOp.name}", end="(")

        for i in range(len(op_list)):
            for ops in op_list[i].params.data:
                self.print_attr_constraint(ops)
                self._print(", ", end='') if i != len(op_list) - 1 else None
        self._print(")")

    def print_dialect_definition(self, di: DialectOp):
        self._print(f"  {DialectOp.name} {di.dialect_name.data} {{")

        di.walk(lambda type: self.print_type_definition(type)
                if isinstance(type, TypeOp) else None)

        di.walk(lambda op: self.print_operation_definition(op)
                if isinstance(op, OperationOp) else None)
        self._print("  }}")



from __future__ import annotations

from typing import Annotated

from xdsl.dialects.builtin import AnyIntegerAttr, IndexType, IntegerAttr
from xdsl.ir import Operation, SSAValue, Block, Region, Dialect
from xdsl.irdl import (OpAttr, VarOpResult, irdl_op_definition, VarOperand,
                       AnyAttr)


@irdl_op_definition
class For(Operation):
    name: str = "affine.for"

    arguments: Annotated[VarOperand, AnyAttr()]
    res: Annotated[VarOpResult, AnyAttr()]

    # TODO the bounds are in fact affine_maps
    # TODO support dynamic bounds as soon as maps are here
    lower_bound: OpAttr[AnyIntegerAttr]
    upper_bound: OpAttr[AnyIntegerAttr]
    step: OpAttr[AnyIntegerAttr]

    body: Region

    def verify_(self) -> None:
        if len(self.operands) != len(self.results):
            raise Exception("Expected the same amount of operands and results")

        operand_types = [SSAValue.get(op).typ for op in self.operands]
        if (operand_types != [res.typ for res in self.results]):
            raise Exception(
                "Expected all operands and result pairs to have matching types"
            )

        entry_block: Block = self.body.blocks[0]
        block_arg_types = [IndexType()] + operand_types
        arg_types = [arg.typ for arg in entry_block.args]
        if block_arg_types != arg_types:
            raise Exception(
                "Expected BlockArguments to have the same types as the operands"
            )

    @staticmethod
    def from_region(operands: list[Operation | SSAValue],
                    lower_bound: int | AnyIntegerAttr,
                    upper_bound: int | AnyIntegerAttr,
                    region: Region,
                    step: int | AnyIntegerAttr = 1) -> For:
        result_types = [SSAValue.get(op).typ for op in operands]
        attributes = {
            "lower_bound": IntegerAttr.from_index_int_value(lower_bound),
            "upper_bound": IntegerAttr.from_index_int_value(upper_bound),
            "step": IntegerAttr.from_index_int_value(step),
        }
        return For.build(operands=[[operand for operand in operands]],
                         result_types=[result_types],
                         attributes=attributes,
                         regions=[region])

    @staticmethod
    def from_callable(operands: list[Operation | SSAValue],
                      lower_bound: int | AnyIntegerAttr,
                      upper_bound: int | AnyIntegerAttr,
                      body: Block.BlockCallback,
                      step: int | AnyIntegerAttr = 1) -> For:
        arg_types = [IndexType()] + [SSAValue.get(op).typ for op in operands]
        return For.from_region(
            operands, lower_bound, upper_bound,
            Region.from_block_list([Block.from_callable(arg_types, body)]),
            step)


@irdl_op_definition
class Yield(Operation):
    name: str = "affine.yield"
    arguments: Annotated[VarOperand, AnyAttr()]

    @staticmethod
    def get(*operands: SSAValue | Operation) -> Yield:
        return Yield.create(
            operands=[SSAValue.get(operand) for operand in operands])


Affine = Dialect([For, Yield], [])



from inspect import isclass
from types import UnionType
from typing import (Annotated, Any, TypeGuard, TypeVar, Union, cast, get_args,
                    get_origin)

_T = TypeVar("_T")


def is_satisfying_hint(arg: Any, hint: type[_T]) -> TypeGuard[_T]:
    """
    Check if `arg` is of the type described by `hint`.
    For now, only lists, tuples, sets, dictionaries, unions,
    and non-generic classes are supported for type hints.
    """
    if hint is Any:
        return True

    # get_origin checks that hint is not a parametrized generic
    if isclass(hint) and (get_origin(hint) is None):
        return isinstance(arg, hint)

    if get_origin(hint) is list:
        if not isinstance(arg, list):
            return False
        arg_list: list[Any] = cast(list[Any], arg)
        elem_hint, = get_args(hint)
        return all(is_satisfying_hint(elem, elem_hint) for elem in arg_list)

    if get_origin(hint) is dict:
        if not isinstance(arg, dict):
            return False
        arg_dict: dict[Any, Any] = cast(dict[Any, Any], arg)
        key_hint, value_hint = get_args(hint)
        return all(
            is_satisfying_hint(key, key_hint)
            and is_satisfying_hint(value, value_hint)
            for key, value in arg_dict.items())

    if get_origin(hint) in [Union, UnionType]:
        return any(
            is_satisfying_hint(arg, union_arg) for union_arg in get_args(hint))

    raise ValueError(f"is_satisfying_hint: unsupported type hint '{hint}'")


annotated_type = type(Annotated[int, 0])
"""This is the type of an Annotated object."""



from __future__ import annotations
from typing import Annotated, List, Union

from xdsl.dialects.builtin import IntegerType
from xdsl.ir import SSAValue, Operation, Block, Dialect
from xdsl.irdl import (irdl_op_definition, VarOperand, AnyAttr, Operand,
                       AttrSizedOperandSegments)


@irdl_op_definition
class Branch(Operation):
    name: str = "cf.br"

    arguments: Annotated[VarOperand, AnyAttr()]

    @staticmethod
    def get(block: Block, *ops: Union[Operation, SSAValue]) -> Branch:
        return Branch.build(operands=[[op for op in ops]], successors=[block])


@irdl_op_definition
class ConditionalBranch(Operation):
    name: str = "cf.cond_br"

    then: Annotated[Operand, IntegerType.from_width(1)]
    then_arguments: Annotated[VarOperand, AnyAttr()]
    else_arguments: Annotated[VarOperand, AnyAttr()]

    irdl_options = [AttrSizedOperandSegments()]

    @staticmethod
    def get(cond: Union[Operation, SSAValue], then_block: Block,
            then_ops: List[Union[Operation, SSAValue]], else_block: Block,
            else_ops: List[Union[Operation, SSAValue]]) -> ConditionalBranch:
        return ConditionalBranch.build(operands=[cond, then_ops, else_ops],
                                       successors=[then_block, else_block])


Cf = Dialect([Branch, ConditionalBranch], [])



from dataclasses import dataclass


@dataclass
class FrontendProgramException(Exception):
    """
    Exception type used when something goes wrong with `FrontendProgram`.
    """

    msg: str

    def __init__(self, msg: str):
        super().__init__()
        self.msg = msg

    def __str__(self) -> str:
        return f"{self.msg}"


@dataclass
class CodeGenerationException(FrontendProgramException):
    """
    Exception type used when xDSL code generation fails. Should be used for
    user-facing errors, e.g. unsupported functionality or failed type checks.
    """

    line: int
    col: int

    def __init__(self, line: int, col: int, msg: str):
        super().__init__(msg)
        self.line = line
        self.col = col

    def __str__(self) -> str:
        return f"Code generation exception at {self.line}:{self.col}. {self.msg}"



from __future__ import annotations
from typing import TYPE_CHECKING, Annotated

from xdsl.ir import (MLIRType, ParametrizedAttribute, Attribute, Dialect,
                     Operation)
from xdsl.irdl import (OpAttr, Operand, ParameterDef, AnyAttr,
                       irdl_attr_definition, irdl_op_definition, builder)
from xdsl.ir import OpResult, Attribute
from xdsl.dialects.builtin import AnyArrayAttr, StringAttr, ArrayOfConstraint, ArrayAttr

if TYPE_CHECKING:
    from xdsl.parser import BaseParser
    from xdsl.printer import Printer


@irdl_attr_definition
class LLVMStructType(ParametrizedAttribute, MLIRType):
    name = "llvm.struct"

    # An empty string refers to a struct without a name.
    struct_name: ParameterDef[StringAttr]
    types: ParameterDef[ArrayAttr[Attribute]]

    # TODO: Add this parameter once xDSL supports the necessary capabilities.
    #  bitmask = ParameterDef(StringAttr)

    @staticmethod
    @builder
    def from_type_list(types: list[Attribute]) -> LLVMStructType:
        return LLVMStructType(
            [StringAttr.from_str(""),
             ArrayAttr.from_list(types)])

    def print_parameters(self, printer: Printer) -> None:
        assert self.struct_name.data == ""
        printer.print("<(")
        printer.print_list(self.types.data, printer.print_attribute)
        printer.print(")>")

    @staticmethod
    def parse_parameters(parser: BaseParser) -> list[Attribute]:
        parser.parse_characters("<(", "LLVM Struct must start with `<(`")
        params = parser.parse_list_of(
            parser.try_parse_type,
            "Malformed LLVM struct, expected attribute definition here!")
        parser.parse_characters(
            ")>", "Unexpected input, expected end of LLVM struct!")
        return [StringAttr.from_str(""), ArrayAttr.from_list(params)]


@irdl_op_definition
class LLVMExtractValue(Operation):
    name = "llvm.extractvalue"

    position: OpAttr[AnyArrayAttr]
    container: Annotated[Operand, AnyAttr()]

    res: Annotated[OpResult, AnyAttr()]


@irdl_op_definition
class LLVMInsertValue(Operation):
    name = "llvm.insertvalue"

    position: OpAttr[AnyArrayAttr]
    container: Annotated[Operand, AnyAttr()]
    value: Annotated[Operand, AnyAttr()]

    res: Annotated[OpResult, AnyAttr()]


@irdl_op_definition
class LLVMMLIRUndef(Operation):
    name = "llvm.mlir.undef"

    res: Annotated[OpResult, AnyAttr()]


LLVM = Dialect([LLVMExtractValue, LLVMInsertValue, LLVMMLIRUndef],
               [LLVMStructType])
